"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
import sys
import tempfile
import torch
from .state import AcceleratorState, PartialState
from .utils import (PrecisionType, PrepareForLaunch, are_libraries_initialized, check_cuda_p2p_ib_support, get_gpu_info, is_mps_available, is_torch_version, patch_environment)
from .utils.constants import ELASTIC_LOG_LINE_PREFIX_TEMPLATE_PYTORCH_VERSION
def test_launch(): _ = PartialState()
def notebook_launcher(function, args=(), num_processes=None, mixed_precision="no", use_port="29500", master_addr="127.0.0.1", node_rank=0, num_nodes=1,
rdzv_backend="static", rdzv_endpoint="", rdzv_conf=None, rdzv_id="none", max_restarts=0, monitor_interval=0.1, log_line_prefix_template=None):
    in_colab = False
    in_kaggle = False
    if any(key.startswith("KAGGLE") for key in os.environ.keys()): in_kaggle = True
    elif "IPython" in sys.modules: in_colab = "google.colab" in str(sys.modules["IPython"].get_ipython())
    try: mixed_precision = PrecisionType(mixed_precision.lower())
    except ValueError: raise ValueError(f"Unknown mixed_precision mode: {args.mixed_precision.lower()}. Choose between {PrecisionType.list()}.")
    if (in_colab or in_kaggle) and (os.environ.get("TPU_NAME", None) is not None):
        import torch_xla.distributed.xla_multiprocessing as xmp
        if len(AcceleratorState._shared_state) > 0: raise ValueError("To train on TPU in Colab or Kaggle Kernel, the `Accelerator` should only be initialized inside your training function. Restart your notebook and make sure no cells initializes an `Accelerator`.")
        if num_processes is None: num_processes = 8
        launcher = PrepareForLaunch(function, distributed_type="XLA")
        xmp.spawn(launcher, args=args, nprocs=num_processes, start_method="fork")
    elif in_colab and get_gpu_info()[1] < 2: function(*args)
    else:
        if num_processes is None: raise ValueError("You have to specify the number of GPUs you would like to use, add `num_processes=...` to your call.")
        if node_rank >= num_nodes: raise ValueError("The node_rank must be less than the number of nodes.")
        if num_processes > 1:
            from torch.distributed.launcher.api import LaunchConfig, elastic_launch
            from torch.multiprocessing import start_processes
            from torch.multiprocessing.spawn import ProcessRaisedException
            if len(AcceleratorState._shared_state) > 0: raise ValueError("To launch a multi-GPU training from your notebook, the `Accelerator` should only be initialized inside your training function. Restart your notebook and make sure no cells initializes an `Accelerator`.")
            problematic_imports = are_libraries_initialized("sapiens_machine")
            if len(problematic_imports) > 0:
                err = ("Could not start distributed process. Libraries known to initialize CUDA upon import have been imported already. Please keep these imports inside your training function to try and help with this:")
                for lib_name in problematic_imports: err += f"\n\t* `{lib_name}`"
                raise RuntimeError(err)
            patched_env = dict(nproc=num_processes, node_rank=node_rank, world_size=num_nodes * num_processes, master_addr=master_addr, master_port=use_port, mixed_precision=mixed_precision)
            if not check_cuda_p2p_ib_support():
                patched_env["nccl_p2p_disable"] = "1"
                patched_env["nccl_ib_disable"] = "1"
            with patch_environment(**patched_env):
                if os.environ.get("SAPIENS_ACCELERATOR_DEBUG_MODE", "false").lower() == "true":
                    launcher = PrepareForLaunch(test_launch, distributed_type="MULTI_GPU")
                    try: start_processes(launcher, args=(), nprocs=num_processes, start_method="fork")
                    except ProcessRaisedException as e:
                        err = "An issue was found when verifying a stable environment for the notebook launcher."
                        if "Cannot re-initialize CUDA in forked subprocess" in e.args[0]: raise RuntimeError(f"{err} This likely stems from an outside import causing issues once the `notebook_launcher()` is called. Please review your imports and test them when running the `notebook_launcher()` to identify which one is problematic and causing CUDA to be initialized.") from e
                        else: raise RuntimeError(f"{err} The following error was raised: {e}") from e
                launcher = PrepareForLaunch(function, distributed_type="MULTI_GPU")
                try:
                    if rdzv_conf is None: rdzv_conf = {}
                    if rdzv_backend == "static":
                        rdzv_conf["rank"] = node_rank
                        if not rdzv_endpoint: rdzv_endpoint = f"{master_addr}:{use_port}"
                    launch_config_kwargs = dict(min_nodes=num_nodes, max_nodes=num_nodes, nproc_per_node=num_processes, run_id=rdzv_id, rdzv_endpoint=rdzv_endpoint,
                    rdzv_backend=rdzv_backend, rdzv_configs=rdzv_conf, max_restarts=max_restarts, monitor_interval=monitor_interval, start_method="fork")
                    if is_torch_version(">=", ELASTIC_LOG_LINE_PREFIX_TEMPLATE_PYTORCH_VERSION): launch_config_kwargs["log_line_prefix_template"] = log_line_prefix_template
                    elastic_launch(config=LaunchConfig(**launch_config_kwargs), entrypoint=function)(*args)
                except ProcessRaisedException as e:
                    if "Cannot re-initialize CUDA in forked subprocess" in e.args[0]: raise RuntimeError("CUDA has been initialized before the `notebook_launcher` could create a forked subprocess. This likely stems from an outside import causing issues once the `notebook_launcher()` is called. Please review your imports and test them when running the `notebook_launcher()` to identify which one is problematic and causing CUDA to be initialized.") from e
                    else: raise RuntimeError(f"An issue was found when launching the training: {e}") from e
        else:
            if is_mps_available(): os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
            function(*args)
def debug_launcher(function, args=(), num_processes=2):
    from torch.multiprocessing import start_processes
    with tempfile.NamedTemporaryFile() as tmp_file:
        with patch_environment(world_size=num_processes, master_addr="127.0.0.1", master_port="29500", sapiens_accelerator_mixed_precision="no", sapiens_accelerator_debug_rdv_file=tmp_file.name, sapiens_accelerator_use_cpu="yes"):
            launcher = PrepareForLaunch(function, debug=True)
            start_processes(launcher, args=args, nprocs=num_processes, start_method="fork")
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""
__version__ = "1.0.0"
from .accelerator import Accelerator
from .big_modeling import (cpu_offload, cpu_offload_with_hook, disk_offload, dispatch_model, init_empty_weights, init_on_device, load_checkpoint_and_dispatch)
from .data_loader import skip_first_batches
from .inference import prepare_pippy
from .launchers import debug_launcher, notebook_launcher
from .state import PartialState
from .utils import (AutocastKwargs, DataLoaderConfiguration, DDPCommunicationHookType, DeepSpeedPlugin, DistributedDataParallelKwargs, DistributedType,
FullyShardedDataParallelPlugin, GradScalerKwargs, InitProcessGroupKwargs, ProfileKwargs, find_executable_batch_size, infer_auto_device_map, is_rich_available,
load_checkpoint_in_model, synchronize_rng_states)
if is_rich_available(): from .utils import rich
"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import base64
import json
import os
from copy import deepcopy
from ..optimizer import SapiensAcceleratordOptimizer
from ..scheduler import SapiensAcceleratordScheduler
from .dataclasses import DistributedType
def get_active_deepspeed_plugin(state):
    if state.distributed_type != DistributedType.DEEPSPEED: raise ValueError("Couldn't retrieve the active `DeepSpeedPlugin` as none were enabled. Please make sure that either `Accelerator` is configured for `deepspeed` or make sure that the desired `DeepSpeedPlugin` has been enabled (`AcceleratorState().select_deepspeed_plugin(name)`) before calling this function.")
    if not isinstance(state.deepspeed_plugins, dict): return state.deepspeed_plugins
    return next(plugin for plugin in state.deepspeed_plugins.values() if plugin.selected)
class HfDeepSpeedConfig:
    def __init__(self, config_file_or_dict):
        if isinstance(config_file_or_dict, dict): config = deepcopy(config_file_or_dict)
        elif os.path.exists(config_file_or_dict):
            with open(config_file_or_dict, encoding="utf-8") as f: config = json.load(f)
        else:
            try:
                config_decoded = base64.urlsafe_b64decode(config_file_or_dict).decode("utf-8")
                config = json.loads(config_decoded)
            except (UnicodeDecodeError, AttributeError, ValueError): raise ValueError(f"Expected a string path to an existing deepspeed config, or a dictionary, or a base64 encoded string. Received: {config_file_or_dict}")
        self.config = config
        self.set_stage_and_offload()
    def set_stage_and_offload(self):
        self._stage = self.get_value("zero_optimization.stage", -1)
        self._offload = False
        if self.is_zero2() or self.is_zero3():
            offload_devices_valid = set(["cpu", "nvme"])
            offload_devices = set([self.get_value("zero_optimization.offload_optimizer.device"), self.get_value("zero_optimization.offload_param.device")])
            if len(offload_devices & offload_devices_valid) > 0: self._offload = True
    def find_config_node(self, ds_key_long):
        config = self.config
        nodes = ds_key_long.split(".")
        ds_key = nodes.pop()
        for node in nodes:
            config = config.get(node)
            if config is None: return None, ds_key
        return config, ds_key
    def get_value(self, ds_key_long, default=None):
        config, ds_key = self.find_config_node(ds_key_long)
        if config is None: return default
        return config.get(ds_key, default)
    def del_config_sub_tree(self, ds_key_long, must_exist=False):
        config = self.config
        nodes = ds_key_long.split(".")
        for node in nodes:
            parent_config = config
            config = config.get(node)
            if config is None:
                if must_exist: raise ValueError(f"Can't find {ds_key_long} entry in the config: {self.config}")
                else: return
        if parent_config is not None: parent_config.pop(node)
    def is_true(self, ds_key_long):
        value = self.get_value(ds_key_long)
        return False if value is None else bool(value)
    def is_false(self, ds_key_long):
        value = self.get_value(ds_key_long)
        return False if value is None else not bool(value)
    def is_zero2(self): return self._stage == 2
    def is_zero3(self): return self._stage == 3
    def is_offload(self): return self._offload
class DeepSpeedEngineWrapper:
    def __init__(self, engine): self.engine = engine
    def backward(self, loss, **kwargs):
        self.engine.backward(loss, **kwargs)
        self.engine.step()
class DeepSpeedOptimizerWrapper(SapiensAcceleratordOptimizer):
    def __init__(self, optimizer):
        super().__init__(optimizer, device_placement=False, scaler=None)
        self.__has_overflow__ = hasattr(self.optimizer, "overflow")
    def zero_grad(self, set_to_none=None): pass
    def step(self): pass
    @property
    def step_was_skipped(self):
        if self.__has_overflow__: return self.optimizer.overflow
        return False
class DeepSpeedSchedulerWrapper(SapiensAcceleratordScheduler):
    def __init__(self, scheduler, optimizers): super().__init__(scheduler, optimizers)
    def step(self): pass
class DummyOptim:
    def __init__(self, params, lr=0.001, weight_decay=0, **kwargs):
        self.params = params
        self.lr = lr
        self.weight_decay = weight_decay
        self.kwargs = kwargs
class DummyScheduler:
    def __init__(self, optimizer, total_num_steps=None, warmup_num_steps=0, lr_scheduler_callable=None, **kwargs):
        self.optimizer = optimizer
        self.total_num_steps = total_num_steps
        self.warmup_num_steps = warmup_num_steps
        self.lr_scheduler_callable = lr_scheduler_callable
        self.kwargs = kwargs
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
from ...utils import (ComputeEnvironment, DistributedType, is_deepspeed_available, is_fp8_available, is_mlu_available, is_mps_available, is_msamp_available,
is_musa_available, is_npu_available, is_transformer_engine_available, is_transformers_available, is_xpu_available)
from ...utils.constants import (DEEPSPEED_MULTINODE_LAUNCHERS, FSDP_AUTO_WRAP_POLICY, FSDP_BACKWARD_PREFETCH, FSDP_SHARDING_STRATEGY, FSDP_STATE_DICT_TYPE, TORCH_DYNAMO_MODES)
from .config_args import ClusterConfig
from .config_utils import (DYNAMO_BACKENDS, _ask_field, _ask_options, _convert_distributed_mode, _convert_dynamo_backend, _convert_fp8_backend, _convert_mixed_precision, _convert_yes_no_to_bool)
def get_cluster_input():
    distributed_type = _ask_options("Which type of machine are you using?", ["No distributed training", "multi-CPU", "multi-XPU", "multi-GPU", "multi-NPU", "multi-MLU", "multi-MUSA", "TPU"], _convert_distributed_mode)
    machine_rank = 0
    num_machines = 1
    num_processes = 1
    gpu_ids = None
    main_process_ip = None
    main_process_port = None
    rdzv_backend = "static"
    same_network = True
    debug = False
    if distributed_type in [DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.MULTI_XPU, DistributedType.MULTI_CPU]:
        num_machines = _ask_field("How many different machines will you use (use more than 1 for multi-node training)? [1]: ", int, default=1)
        if num_machines > 1:
            machine_rank = _ask_options("What is the rank of this machine?", list(range(num_machines)), int)
            main_process_ip = _ask_field("What is the IP address of the machine that will host the main process? ")
            main_process_port = _ask_field("What is the port you will use to communicate with the main process? ", int)
            same_network = _ask_field("Are all the machines on the same local network? Answer `no` if nodes are on the cloud and/or on different network hosts [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            if not same_network: rdzv_backend = _ask_field("What rendezvous backend will you use? ('static', 'c10d', ...): ", default="static")
        debug = _ask_field("Should distributed operations be checked while running for errors? This can avoid timeout issues but will be slower. [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    if distributed_type == DistributedType.NO: use_cpu = _ask_field("Do you want to run your training on CPU only (even if a GPU / Apple Silicon / Ascend NPU device is available)? [yes/NO]:", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    elif distributed_type == DistributedType.MULTI_CPU: use_cpu = True
    else: use_cpu = False
    ipex_config = {}
    mpirun_config = {}
    if use_cpu:
        ipex_config["ipex"] = _ask_field("Do you want to use Intel PyTorch Extension (IPEX) to speed up training on CPU? [yes/NO]:", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if distributed_type == DistributedType.MULTI_CPU:
            use_mpirun = _ask_field("Do you want sapiens_accelerator to launch mpirun? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            if use_mpirun:
                mpirun_hostfile = _ask_field("Please enter the path to the hostfile to use with mpirun [~/hostfile]: ", str, default="~/hostfile")
                mpirun_config["mpirun_hostfile"] = os.path.expanduser(mpirun_hostfile.strip())
                mpirun_config["mpirun_ccl"] = _ask_field("Enter the number of oneCCL worker threads [1]: ", default=1)
    if (not use_cpu and is_xpu_available() and distributed_type not in [DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU,
    DistributedType.XLA, DistributedType.MULTI_MUSA]):
        ipex_config["use_xpu"] = _ask_field("Do you want to use XPU plugin to speed up training on XPU? [yes/NO]:", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    dynamo_config = {}
    use_dynamo = _ask_field("Do you wish to optimize your script with torch dynamo?[yes/NO]:", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    if use_dynamo:
        prefix = "dynamo_"
        dynamo_config[prefix + "backend"] = _ask_options("Which dynamo backend would you like to use?", [x.lower() for x in DYNAMO_BACKENDS], _convert_dynamo_backend, default=2)
        use_custom_options = _ask_field("Do you want to customize the defaults sent to torch.compile? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if use_custom_options:
            dynamo_config[prefix + "mode"] = _ask_options("Which mode do you want to use?", TORCH_DYNAMO_MODES, lambda x: TORCH_DYNAMO_MODES[int(x)], default=0)
            dynamo_config[prefix + "use_fullgraph"] = _ask_field("Do you want the fullgraph mode or it is ok to break model into several subgraphs? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            dynamo_config[prefix + "use_dynamic"] = _ask_field("Do you want to enable dynamic shape tracing? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    use_mps = not use_cpu and is_mps_available()
    deepspeed_config = {}
    if (distributed_type in [DistributedType.MULTI_GPU, DistributedType.MULTI_XPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU,
    DistributedType.MULTI_MUSA, DistributedType.NO] and not use_mps):
        use_deepspeed = _ask_field("Do you want to use DeepSpeed? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if use_deepspeed:
            distributed_type = DistributedType.DEEPSPEED
            assert (is_deepspeed_available()), "DeepSpeed is not installed => run `pip3 install deepspeed` or build it from source"
        if distributed_type == DistributedType.DEEPSPEED:
            use_deepspeed_config = _ask_field("Do you want to specify a json file to a DeepSpeed config? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            if use_deepspeed_config: deepspeed_config["deepspeed_config_file"] = _ask_field("Please enter the path to the json DeepSpeed config file: ", str, default="none")
            else:
                deepspeed_config["zero_stage"] = _ask_options("What should be your DeepSpeed's ZeRO optimization stage?", [0, 1, 2, 3], int, default=2)
                deepspeed_devices = ["none", "cpu", "nvme"]
                if deepspeed_config["zero_stage"] >= 2:
                    deepspeed_config["offload_optimizer_device"] = _ask_options("Where to offload optimizer states?", deepspeed_devices, lambda x: deepspeed_devices[int(x)])
                    deepspeed_config["offload_param_device"] = _ask_options("Where to offload parameters?", deepspeed_devices, lambda x: deepspeed_devices[int(x)])
                    if deepspeed_config["offload_param_device"] == "nvme": deepspeed_config["offload_param_nvme_path"] = _ask_field("Nvme Path to offload parameters?", str, default="/nvme")
                    if deepspeed_config["offload_optimizer_device"] == "nvme": deepspeed_config["offload_optimizer_nvme_path"] = _ask_field("Nvme Path to offload optimizer states?", str, default="/nvme")
                deepspeed_config["gradient_accumulation_steps"] = _ask_field("How many gradient accumulation steps you're passing in your script? [1]: ", int, default=1)
                use_gradient_clipping = _ask_field("Do you want to use gradient clipping? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
                if use_gradient_clipping: deepspeed_config["gradient_clipping"] = _ask_field("What is the gradient clipping value? [1.0]: ", float, default=1.0)
                if deepspeed_config["zero_stage"] == 3: deepspeed_config["zero3_save_16bit_model"] = _ask_field("Do you want to save 16-bit model weights when using ZeRO Stage-3? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            deepspeed_config["zero3_init_flag"] = _ask_field("Do you want to enable `deepspeed.zero.Init` when using ZeRO Stage-3 for constructing massive models? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            if deepspeed_config["zero3_init_flag"]:
                if not is_transformers_available(): raise Exception("When `zero3_init_flag` is set, it requires Transformers to be installed. Please run `pip3 install transformers`.")
            use_moe = _ask_field("Do you want to enable Mixture-of-Experts training (MoE)? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            if use_moe: deepspeed_config["deepspeed_moe_layer_cls_names"] = _ask_field("Specify the comma-separated list of transformers MoE layer class names (case-sensitive), e.g : `MixtralSparseMoeBlock`, `Qwen2MoeSparseMoeBlock`, `JetMoEAttention,JetMoEBlock` ... : ", str)
            if num_machines > 1:
                launcher_query = "Which Type of launcher do you want to use?"
                deepspeed_config["deepspeed_multinode_launcher"] = _ask_options(launcher_query, DEEPSPEED_MULTINODE_LAUNCHERS, lambda x: DEEPSPEED_MULTINODE_LAUNCHERS[int(x)])
                if deepspeed_config["deepspeed_multinode_launcher"] != DEEPSPEED_MULTINODE_LAUNCHERS[1]:
                    deepspeed_config["deepspeed_hostfile"] = _ask_field("DeepSpeed configures multi-node compute resources with hostfile. Each row is of the format `hostname slots=[num_gpus]`, e.g., `localhost slots=2`; for more information please refer official [documentation] (https://www.deepspeed.ai/getting-started/#resource-configuration-multi-node). Please specify the location of hostfile: ", str)
                    is_exclusion_filter = _ask_field("Do you want to specify exclusion filter string? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
                    if is_exclusion_filter: deepspeed_config["deepspeed_exclusion_filter"] = _ask_field("DeepSpeed exclusion filter string: ", str)
                    is_inclusion_filter = _ask_field("Do you want to specify inclusion filter string? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
                    if is_inclusion_filter: deepspeed_config["deepspeed_inclusion_filter"] = _ask_field("DeepSpeed inclusion filter string: ", str)
    fsdp_config = {}
    if distributed_type in [DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_XPU]:
        use_fsdp = _ask_field("Do you want to use FullyShardedDataParallel? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if use_fsdp: distributed_type = DistributedType.FSDP
        if distributed_type == DistributedType.FSDP:
            sharding_strategy_query = "What should be your sharding strategy?"
            fsdp_config["fsdp_sharding_strategy"] = _ask_options(sharding_strategy_query, FSDP_SHARDING_STRATEGY, lambda x: FSDP_SHARDING_STRATEGY[int(x)])
            fsdp_config["fsdp_offload_params"] = _ask_field("Do you want to offload parameters and gradients to CPU? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            fsdp_wrap_query = "What should be your auto wrap policy?"
            fsdp_config["fsdp_auto_wrap_policy"] = _ask_options(fsdp_wrap_query, FSDP_AUTO_WRAP_POLICY, lambda x: FSDP_AUTO_WRAP_POLICY[int(x)])
            if fsdp_config["fsdp_auto_wrap_policy"] == FSDP_AUTO_WRAP_POLICY[0]:
                use_no_split_modules = _ask_field("Do you want to use the model's `_no_split_modules` to wrap. Only applicable for Sapiens Transformers [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
                if not use_no_split_modules: fsdp_config["fsdp_transformer_layer_cls_to_wrap"] = _ask_field("Specify the comma-separated list of transformer layer class names (case-sensitive) to wrap ,e.g, : `BertLayer`, `GPTJBlock`, `T5Block`, `BertLayer,BertEmbeddings,BertSelfOutput` ...? : ", str)
            elif fsdp_config["fsdp_auto_wrap_policy"] == FSDP_AUTO_WRAP_POLICY[1]: fsdp_config["fsdp_min_num_params"] = _ask_field("What should be your FSDP's minimum number of parameters for Default Auto Wrapping Policy? [1e8]: ", int, default=100000000)
            fsdp_backward_prefetch_query = "What should be your FSDP's backward prefetch policy?"
            fsdp_config["fsdp_backward_prefetch"] = _ask_options(fsdp_backward_prefetch_query, FSDP_BACKWARD_PREFETCH, lambda x: FSDP_BACKWARD_PREFETCH[int(x)])
            fsdp_state_dict_type_query = "What should be your FSDP's state dict type?"
            fsdp_config["fsdp_state_dict_type"] = _ask_options(fsdp_state_dict_type_query, FSDP_STATE_DICT_TYPE, lambda x: FSDP_STATE_DICT_TYPE[int(x)], default=2)
            fsdp_config["fsdp_forward_prefetch"] = _ask_field("Do you want to enable FSDP's forward prefetch policy? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            fsdp_config["fsdp_use_orig_params"] = _ask_field("Do you want to enable FSDP's `use_orig_params` feature? [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            fsdp_config["fsdp_cpu_ram_efficient_loading"] = _ask_field("Do you want to enable CPU RAM efficient model loading? Only applicable for Sapiens Transformers models. [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            if fsdp_config["fsdp_cpu_ram_efficient_loading"]: fsdp_config["fsdp_sync_module_states"] = True
            else: fsdp_config["fsdp_sync_module_states"] = _ask_field("Do you want each individually wrapped FSDP unit to broadcast module parameters from rank 0 at the start? [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            fsdp_config["fsdp_activation_checkpointing"] = _ask_field("Do you want to enable FSDP activation checkpointing? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    megatron_lm_config = {}
    if distributed_type in [DistributedType.MULTI_GPU]:
        use_megatron_lm = _ask_field("Do you want to use Megatron-LM ? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if use_megatron_lm: distributed_type = DistributedType.MEGATRON_LM
        if distributed_type == DistributedType.MEGATRON_LM:
            prefix = "megatron_lm_"
            megatron_lm_config[prefix + "tp_degree"] = _ask_field("What is the Tensor Parallelism degree/size? [1]:", int, default=1, error_message="Please enter an integer.")
            if megatron_lm_config[prefix + "tp_degree"] > 1: megatron_lm_config[prefix + "sequence_parallelism"] = _ask_field("Do you want to enable Sequence Parallelism? [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            megatron_lm_config[prefix + "pp_degree"] = _ask_field("What is the Pipeline Parallelism degree/size? [1]:", int, default=1, error_message="Please enter an integer.")
            if megatron_lm_config[prefix + "pp_degree"] > 1: megatron_lm_config[prefix + "num_micro_batches"] = _ask_field("What is the number of micro-batches? [1]:", int, default=1, error_message="Please enter an integer.")
            megatron_lm_config[prefix + "recompute_activations"] = _ask_field("Do you want to enable selective activation recomputation? [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            megatron_lm_config[prefix + "use_distributed_optimizer"] = _ask_field("Do you want to use distributed optimizer which shards optimizer state and gradients across data parallel ranks? [YES/no]: ", _convert_yes_no_to_bool, default=True, error_message="Please enter yes or no.")
            megatron_lm_config[prefix + "gradient_clipping"] = _ask_field("What is the gradient clipping value based on global L2 Norm (0 to disable)? [1.0]: ", float, default=1.0)
    tpu_commands = None
    tpu_command_file = None
    tpu_downcast_bf16 = "no"
    tpu_env = []
    tpu_name = None
    tpu_vm = None
    tpu_zone = None
    tpu_use_sudo = False
    tpu_use_cluster = False
    if distributed_type in [DistributedType.MULTI_CPU, DistributedType.MULTI_XPU, DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.XLA]:
        machine_type = str(distributed_type).split(".")[1].replace("MULTI_", "")
        if machine_type == "TPU": machine_type += " cores"
        elif machine_type == "CPU": machine_type = "processes"
        else: machine_type += "(s)"
        num_processes = _ask_field(f"How many {machine_type} should be used for distributed training? [1]:", int, default=1, error_message="Please enter an integer.")
    elif distributed_type in [DistributedType.FSDP, DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM]: num_processes = _ask_field("How many GPU(s) should be used for distributed training? [1]:", int, default=1, error_message="Please enter an integer.")
    else: num_processes = 1
    if (distributed_type == DistributedType.MULTI_GPU) and (num_machines == 1) and (num_processes == 1): raise ValueError(f"Specified distributed type {distributed_type} but only using 1 GPU on a single machine. Please select `No distributed training` for the type of machine you are using.")
    if (distributed_type in [DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.MULTI_XPU, DistributedType.NO] and not use_cpu and not use_mps):
        if is_npu_available(): machine_type = "NPU(s)"
        elif is_mlu_available(): machine_type = "MLU(s)"
        elif is_musa_available(): machine_type = "MUSA(s)"
        else: machine_type = "GPU(s)"
        gpu_ids = _ask_field(f"What {machine_type} (by id) should be used for training on this machine as a comma-seperated list? [all]:", default="all")
    enable_cpu_affinity = False
    if distributed_type in (DistributedType.NO, DistributedType.MULTI_GPU) and not use_cpu and not use_mps: enable_cpu_affinity = _ask_field("Would you like to enable numa efficiency? (Currently only supported on NVIDIA hardware). [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
    fp8_config = None
    if distributed_type == DistributedType.XLA:
        mixed_precision = "no"
        main_training_function = _ask_field("What is the name of the function in your script that should be launched in all parallel scripts? [main]: ", default="main")
        tpu_use_cluster = _ask_field("Are you using a TPU cluster? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
        if tpu_use_cluster:
            tpu_name = _ask_field("What is the name of your TPU cluster? ", default=None, error_message="Please enter the name of your TPU cluster.")
            tpu_zone = _ask_field("What is the zone of your TPU cluster? ", default=None, error_message="Please enter the zone of your TPU cluster.")
            tpu_use_sudo = _ask_field("To run a python script in a TPU pod, should `sudo` be used? [yes/NO]: ", default=False, error_message="Please enter yes or no.")
            run_commands = _ask_field("Do you have code you wish to run on startup in each pod? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            if run_commands:
                use_command_file = _ask_field("Is this code located in a bash script? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
                if use_command_file:
                    tpu_command_file = _ask_field("What is the path to your bash script? ", default=None, error_message="Please enter the path to your bash script.")
                    tpu_command_file = os.path.abspath(tpu_command_file)
                else:
                    print("Please enter each command seperately you wish to run on startup in each pod.")
                    tpu_commands = []
                    another_command = True
                    while another_command:
                        tpu_commands.append(_ask_field("Please enter a single command to be ran ", default=None, error_message="Please enter the commands you wish to run on startup in each pod as a single string."))
                        another_command = _ask_field("Do you wish to add another command? [yes/NO]: ", _convert_yes_no_to_bool, default=False, error_message="Please enter yes or no.")
            tpu_vm = _ask_field("If not using an instance group, what are the names of the Compute VM instances to be used, seperated by a comma: ", default="").split(",")
            tpu_env = _ask_field("What environment variables do you wish to set in each pod, seperated by a comma: ", default="").split(",")
    else:
        main_training_function = "main"
        if distributed_type == DistributedType.DEEPSPEED and use_deepspeed_config: mixed_precision = None
        else:
            mixed_precision = _ask_options("Do you wish to use mixed precision?", ["no", "fp16", "bf16", "fp8"], _convert_mixed_precision)
            if mixed_precision == "fp8":
                if not is_fp8_available(): raise ValueError("FP8 (either Transformer Engine or MSAMP) is not installed on this machine.")
                fp8_config = {}
                fp8_config["backend"] = _ask_options("Which FP8 backend do you want to use?", ["te", "msamp"], _convert_fp8_backend)
                if fp8_config["backend"] == "TE":
                    if not is_transformer_engine_available(): raise ValueError("TransformersEngine was selected, but it is not installed on this machine.")
                    fp8_config["use_autocast_during_eval"] = _ask_field("Do you want to use FP8 autocast during eval mode? Generally better metrics are found when this is disabled [yes/NO]: ", _convert_yes_no_to_bool, default=False)
                    fp8_config["margin"] = _ask_field("What margin should be used for gradient scaling? [0]: ", int, default=0)
                    fp8_config["interval"] = _ask_field("What interval should be used for for how often the scaling factor is recomputed? [1]: ", int, default=1)
                    fp8_config["fp8_format"] = _ask_options("Which weight format should be used?", ["HYBRID", "E4M3"], lambda x: "HYBRID" if x == 0 else "E4M3", default=0)
                    fp8_config["amax_history_length"] = _ask_field("What length of history should be used for the amax scaling factor computation? [1024]: ", int, default=1024)
                    fp8_config["amax_compute_algorithm"] = _ask_options("Which algorithm should be used for the amax scaling factor computation?", ["max", "most_recent"], lambda x: "max" if x == 0 else "most_recent", default=0)
                    fp8_config["override_linear_precision"] = _ask_field("Do you want to to execute `fprop`, `dgrad`, and `wgrad` GEMMS in higher precision? [yes/NO]: ", _convert_yes_no_to_bool, default=False)
                    if fp8_config["override_linear_precision"]:
                        fprop = _ask_field("Should `fprop` be executed in higher precision? [yes/NO]: ", _convert_yes_no_to_bool, default=False)
                        dgrad = _ask_field("Should `dgrad` be executed in higher precision? [yes/NO]: ", _convert_yes_no_to_bool, default=False)
                        wgrad = _ask_field("Should `wgrad` be executed in higher precision? [yes/NO]: ", _convert_yes_no_to_bool, default=False)
                        fp8_config["override_linear_precision"] = (fprop, dgrad, wgrad)
                elif fp8_config["backend"] == "MSAMP":
                    if not is_msamp_available(): raise ValueError("MSAMP was selected, but it is not installed on this machine.")
                    fp8_config["optimization_level"] = _ask_options("Which optimization level should be used?", ["O1", "O2"], lambda x: "O1" if x == 0 else "O2", default=1)
    if use_dynamo and mixed_precision == "no" and not use_cpu: print("Torch dynamo used without mixed precision requires TF32 to be efficient. SapiensAccelerator will enable it by default when launching your scripts.")
    if distributed_type == DistributedType.XLA and mixed_precision == "bf16": tpu_downcast_bf16 = _ask_field("Should `torch.float` be cast as `bfloat16` and `torch.double` remain `float32` on TPUs?", default="no")
    return ClusterConfig(compute_environment=ComputeEnvironment.LOCAL_MACHINE, distributed_type=distributed_type, num_processes=num_processes, gpu_ids=gpu_ids,
    mixed_precision=mixed_precision, downcast_bf16=tpu_downcast_bf16, machine_rank=machine_rank, num_machines=num_machines, main_process_ip=main_process_ip,
    main_process_port=main_process_port, main_training_function=main_training_function, fp8_config=fp8_config, deepspeed_config=deepspeed_config, fsdp_config=fsdp_config,
    megatron_lm_config=megatron_lm_config, ipex_config=ipex_config, mpirun_config=mpirun_config, use_cpu=use_cpu, rdzv_backend=rdzv_backend, same_network=same_network,
    commands=tpu_commands, command_file=tpu_command_file, tpu_env=tpu_env, tpu_name=tpu_name, tpu_vm=tpu_vm, tpu_zone=tpu_zone, tpu_use_sudo=tpu_use_sudo,
    tpu_use_cluster=tpu_use_cluster, dynamo_config=dynamo_config, debug=debug, enable_cpu_affinity=enable_cpu_affinity)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
import time
from multiprocessing import Queue
from pytest import mark, raises
from torch.distributed.elastic.multiprocessing.errors import ChildFailedError
from sapiens_accelerator import PartialState, notebook_launcher
from sapiens_accelerator.test_utils import require_sapiens
from sapiens_accelerator.utils import is_sapiens_available
def basic_function(): pass
def tough_nut_function(queue: Queue):
    if queue.empty(): return
    trial = queue.get()
    if trial > 0:
        queue.put(trial - 1)
        raise RuntimeError("The nut hasn't cracked yet! Try again.")
def bipolar_sleep_function(sleep_sec: int):
    state = PartialState()
    if state.process_index % 2 == 0: raise RuntimeError("I'm an even process. I don't like to sleep.")
    else: time.sleep(sleep_sec)
NUM_PROCESSES = int(os.environ.get("SAPIENS_ACCELERATOR_NUM_PROCESSES", 1))
def test_can_initialize(): notebook_launcher(basic_function, (), num_processes=NUM_PROCESSES)
@mark.skipif(NUM_PROCESSES < 2, reason="Need at least 2 processes to test static rendezvous backends")
def test_static_rdzv_backend(): notebook_launcher(basic_function, (), num_processes=NUM_PROCESSES, rdzv_backend="static")
@mark.skipif(NUM_PROCESSES < 2, reason="Need at least 2 processes to test c10d rendezvous backends")
def test_c10d_rdzv_backend(): notebook_launcher(basic_function, (), num_processes=NUM_PROCESSES, rdzv_backend="c10d")
@mark.skipif(NUM_PROCESSES < 2, reason="Need at least 2 processes to test fault tolerance")
def test_fault_tolerant(max_restarts: int = 3):
    queue = Queue()
    queue.put(max_restarts)
    notebook_launcher(tough_nut_function, (queue,), num_processes=NUM_PROCESSES, max_restarts=max_restarts)
@mark.skipif(NUM_PROCESSES < 2, reason="Need at least 2 processes to test monitoring")
def test_monitoring(monitor_interval: float = 0.01, sleep_sec: int = 100):
    start_time = time.time()
    with raises(ChildFailedError, match="I'm an even process. I don't like to sleep."): notebook_launcher(bipolar_sleep_function, (sleep_sec,), num_processes=NUM_PROCESSES, monitor_interval=monitor_interval)
    assert time.time() - start_time < sleep_sec, "Monitoring did not stop the process in time."
@require_sapiens
def test_problematic_imports():
    with raises(RuntimeError, match="Please keep these imports"):
        import sapiens_machine as sapiens
        notebook_launcher(basic_function, (), num_processes=NUM_PROCESSES)
def main():
    test_can_initialize()
    test_static_rdzv_backend()
    test_c10d_rdzv_backend()
    test_fault_tolerant()
    test_monitoring()
    if is_sapiens_available(): test_problematic_imports()
    if NUM_PROCESSES > 1: PartialState().destroy_process_group()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import logging
import os
from contextlib import contextmanager
from functools import wraps
from typing import Dict, List, Optional, Union
import torch
import torch.nn as nn
from .hooks import (AlignDevicesHook, CpuOffload, UserCpuOffloadHook, add_hook_to_module, attach_align_device_hook, attach_align_device_hook_on_blocks)
from .utils import (OffloadedWeightsLoader, check_cuda_p2p_ib_support, check_device_map, extract_submodules_state_dict, find_tied_parameters, get_balanced_memory,
infer_auto_device_map, is_mlu_available, is_musa_available, is_npu_available, is_torch_version, is_xpu_available, load_checkpoint_in_model, offload_state_dict,
parse_flag_from_env, retie_parameters)
from .utils.other import recursive_getattr
logger = logging.getLogger(__name__)
@contextmanager
def init_empty_weights(include_buffers: bool = None):
    if include_buffers is None: include_buffers = parse_flag_from_env("SAPIENS_ACCELERATOR_INIT_INCLUDE_BUFFERS", False)
    with init_on_device(torch.device("meta"), include_buffers=include_buffers) as f: yield f
@contextmanager
def init_on_device(device: torch.device, include_buffers: bool = None):
    if include_buffers is None: include_buffers = parse_flag_from_env("SAPIENS_ACCELERATOR_INIT_INCLUDE_BUFFERS", False)
    if is_torch_version(">=", "2.0") and include_buffers:
        with device: yield
        return
    old_register_parameter = nn.Module.register_parameter
    if include_buffers: old_register_buffer = nn.Module.register_buffer
    def register_empty_parameter(module, name, param):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            kwargs = module._parameters[name].__dict__
            kwargs["requires_grad"] = param.requires_grad
            module._parameters[name] = param_cls(module._parameters[name].to(device), **kwargs)
    def register_empty_buffer(module, name, buffer, persistent=True):
        old_register_buffer(module, name, buffer, persistent=persistent)
        if buffer is not None: module._buffers[name] = module._buffers[name].to(device)
    if include_buffers: tensor_constructors_to_patch = {torch_function_name: getattr(torch, torch_function_name) for torch_function_name in ["empty", "zeros", "ones", "full"]}
    else: tensor_constructors_to_patch = {}
    def patch_tensor_constructor(fn):
        def wrapper(*args, **kwargs):
            kwargs["device"] = device
            return fn(*args, **kwargs)
        return wrapper
    try:
        nn.Module.register_parameter = register_empty_parameter
        if include_buffers: nn.Module.register_buffer = register_empty_buffer
        for torch_function_name in tensor_constructors_to_patch.keys(): setattr(torch, torch_function_name, patch_tensor_constructor(getattr(torch, torch_function_name)))
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter
        if include_buffers: nn.Module.register_buffer = old_register_buffer
        for torch_function_name, old_torch_function in tensor_constructors_to_patch.items(): setattr(torch, torch_function_name, old_torch_function)
def cpu_offload(model: nn.Module, execution_device: Optional[torch.device] = None, offload_buffers: bool = False, state_dict: Optional[Dict[str, torch.Tensor]] = None, preload_module_classes: Optional[List[str]] = None):
    if execution_device is None: execution_device = next(iter(model.parameters())).device
    if state_dict is None: state_dict = {n: p.to("cpu") for n, p in model.state_dict().items()}
    add_hook_to_module(model, AlignDevicesHook(io_same_device=True), append=True)
    attach_align_device_hook(model, execution_device=execution_device, offload=True, offload_buffers=offload_buffers, weights_map=state_dict, preload_module_classes=preload_module_classes)
    return model
def cpu_offload_with_hook(model: torch.nn.Module, execution_device: Optional[Union[int, str, torch.device]] = None, prev_module_hook: Optional[UserCpuOffloadHook] = None):
    hook = CpuOffload(execution_device=execution_device, prev_module_hook=prev_module_hook)
    add_hook_to_module(model, hook, append=True)
    user_hook = UserCpuOffloadHook(model, hook)
    return model, user_hook
def disk_offload(model: nn.Module, offload_dir: Union[str, os.PathLike], execution_device: Optional[torch.device] = None, offload_buffers: bool = False, preload_module_classes: Optional[List[str]] = None):
    if not os.path.isdir(offload_dir) or not os.path.isfile(os.path.join(offload_dir, "index.json")): offload_state_dict(offload_dir, model.state_dict())
    if execution_device is None: execution_device = next(iter(model.parameters())).device
    weights_map = OffloadedWeightsLoader(save_folder=offload_dir)
    add_hook_to_module(model, AlignDevicesHook(io_same_device=True), append=True)
    attach_align_device_hook(model, execution_device=execution_device, offload=True, offload_buffers=offload_buffers, weights_map=weights_map, preload_module_classes=preload_module_classes)
    return model
def dispatch_model(model: nn.Module, device_map: Dict[str, Union[str, int, torch.device]], main_device: Optional[torch.device] = None, state_dict: Optional[Dict[str, torch.Tensor]] = None,
offload_dir: Optional[Union[str, os.PathLike]] = None, offload_index: Optional[Dict[str, str]] = None, offload_buffers: bool = False, skip_keys: Optional[Union[str, List[str]]] = None,
preload_module_classes: Optional[List[str]] = None, force_hooks: bool = False):
    check_device_map(model, device_map)
    is_sapiens_quantized = (getattr(model, "is_quantized", False) or getattr(model, "is_loaded_in_8bit", False)) and getattr(model, "quantization_method", "sapiens_machine") == "sapiens_machine"
    if (len(set(device_map.values())) > 1) or is_sapiens_quantized or force_hooks:
        if main_device is None:
            if set(device_map.values()) == {"cpu"} or set(device_map.values()) == {"cpu", "disk"}: main_device = "cpu"
            else: main_device = [d for d in device_map.values() if d not in ["cpu", "disk"]][0]
        if main_device != "cpu":
            cpu_modules = [name for name, device in device_map.items() if device == "cpu"]
            if state_dict is None and len(cpu_modules) > 0: state_dict = extract_submodules_state_dict(model.state_dict(), cpu_modules)
        disk_modules = [name for name, device in device_map.items() if device == "disk"]
        if offload_dir is None and offload_index is None and len(disk_modules) > 0: raise ValueError(f"We need an `offload_dir` to dispatch this model according to this `device_map`, the following submodules need to be offloaded: {', '.join(disk_modules)}.")
        if (len(disk_modules) > 0 and offload_index is None and (not os.path.isdir(offload_dir) or not os.path.isfile(os.path.join(offload_dir, "index.json")))):
            disk_state_dict = extract_submodules_state_dict(model.state_dict(), disk_modules)
            offload_state_dict(offload_dir, disk_state_dict)
        execution_device = {name: main_device if device in ["cpu", "disk"] else device for name, device in device_map.items()}
        execution_device[""] = main_device
        offloaded_devices = ["disk"] if main_device == "cpu" or main_device == "mps" else ["cpu", "disk"]
        offload = {name: device in offloaded_devices for name, device in device_map.items()}
        save_folder = offload_dir if len(disk_modules) > 0 else None
        if state_dict is not None or save_folder is not None or offload_index is not None:
            device = main_device if offload_index is not None else None
            weights_map = OffloadedWeightsLoader(state_dict=state_dict, save_folder=save_folder, index=offload_index, device=device)
        else: weights_map = None
        tied_params = find_tied_parameters(model)
        tied_params_map = {}
        for group in tied_params:
            for param_name in group:
                data_ptr = recursive_getattr(model, param_name).data_ptr()
                tied_params_map[data_ptr] = {}
        attach_align_device_hook_on_blocks(model, execution_device=execution_device, offload=offload, offload_buffers=offload_buffers, weights_map=weights_map,
        skip_keys=skip_keys, preload_module_classes=preload_module_classes, tied_params_map=tied_params_map)
        offloaded_devices_str = " and ".join([device for device in set(device_map.values()) if device in ("cpu", "disk")])
        retie_parameters(model, tied_params)
        def add_warning(fn, model):
            @wraps(fn)
            def wrapper(*args, **kwargs):
                for param in model.parameters():
                    if param.device == torch.device("meta"): raise RuntimeError("You can't move a model that has some modules offloaded to cpu or disk.")
                return fn(*args, **kwargs)
            return wrapper
        model.to = add_warning(model.to, model)
        if is_npu_available(): model.npu = add_warning(model.npu, model)
        elif is_mlu_available(): model.mlu = add_warning(model.mlu, model)
        elif is_musa_available(): model.musa = add_warning(model.musa, model)
        elif is_xpu_available(): model.xpu = add_warning(model.xpu, model)
        else: model.cuda = add_warning(model.cuda, model)
        use_multi_gpu = len([device for device in set(device_map.values()) if device not in ("cpu", "disk")]) > 1
    else:
        device = list(device_map.values())[0]
        if is_npu_available() and isinstance(device, int): device = f"npu:{device}"
        elif is_mlu_available() and isinstance(device, int): device = f"mlu:{device}"
        elif is_musa_available() and isinstance(device, int): device = f"musa:{device}"
        elif is_xpu_available() and isinstance(device, int): device = f"xpu:{device}"
        if device != "disk": model.to(device)
        else: raise ValueError("You are trying to offload the whole model to the disk. Please use the `disk_offload` function instead.")
    model.hf_device_map = dict(device_map)
    return model
def load_checkpoint_and_dispatch(model: nn.Module, checkpoint: Union[str, os.PathLike], device_map: Optional[Union[str, Dict[str, Union[int, str, torch.device]]]] = None,
max_memory: Optional[Dict[Union[int, str], Union[int, str]]] = None, no_split_module_classes: Optional[List[str]] = None, offload_folder: Optional[Union[str, os.PathLike]] = None,
offload_buffers: bool = False, dtype: Optional[Union[str, torch.dtype]] = None, offload_state_dict: Optional[bool] = None, skip_keys: Optional[Union[str, List[str]]] = None,
preload_module_classes: Optional[List[str]] = None, force_hooks: bool = False, strict: bool = False):
    if isinstance(device_map, str) and device_map not in ["auto", "balanced", "balanced_low_0", "sequential"]: raise ValueError("If passing a string for `device_map`, please choose 'auto', 'balanced', 'balanced_low_0' or 'sequential'.")
    if isinstance(device_map, str):
        if device_map != "sequential": max_memory = get_balanced_memory(model, max_memory=max_memory, no_split_module_classes=no_split_module_classes, dtype=dtype, low_zero=(device_map == "balanced_low_0"))
        device_map = infer_auto_device_map(model, max_memory=max_memory, no_split_module_classes=no_split_module_classes, dtype=dtype, offload_buffers=offload_buffers)
    if offload_state_dict is None and device_map is not None and "disk" in device_map.values(): offload_state_dict = True
    load_checkpoint_in_model(model, checkpoint, device_map=device_map, offload_folder=offload_folder, dtype=dtype, offload_state_dict=offload_state_dict,
    offload_buffers=offload_buffers, strict=strict)
    if device_map is None: return model
    return dispatch_model(model, device_map=device_map, offload_dir=offload_folder, offload_buffers=offload_buffers, skip_keys=skip_keys,
    preload_module_classes=preload_module_classes, force_hooks=force_hooks)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import importlib
import importlib.metadata
import os
import warnings
from functools import lru_cache, wraps
import torch
from packaging import version
from packaging.version import parse
from .environment import parse_flag_from_env, str_to_bool
from .versions import compare_versions, is_torch_version
USE_TORCH_XLA = parse_flag_from_env("USE_TORCH_XLA", default=True)
_torch_xla_available = False
if USE_TORCH_XLA:
    try:
        import torch_xla.core.xla_model as xm
        import torch_xla.runtime
        _torch_xla_available = True
    except ImportError: pass
_tpu_available = _torch_xla_available
_torch_distributed_available = torch.distributed.is_available()
def _is_package_available(pkg_name, metadata_name=None):
    package_exists = importlib.util.find_spec(pkg_name) is not None
    if package_exists:
        try:
            _ = importlib.metadata.metadata(pkg_name if metadata_name is None else metadata_name)
            return True
        except importlib.metadata.PackageNotFoundError: return False
def is_torch_distributed_available() -> bool: return _torch_distributed_available
def is_ccl_available():
    try: pass
    except ImportError: pass
    return (importlib.util.find_spec("torch_ccl") is not None or importlib.util.find_spec("oneccl_bindings_for_pytorch") is not None)
def get_ccl_version(): return importlib.metadata.version("oneccl_bind_pt")
def is_import_timer_available(): return _is_package_available("import_timer")
def is_pynvml_available(): return _is_package_available("pynvml") or _is_package_available("pynvml", "nvidia-ml-py")
def is_pytest_available(): return _is_package_available("pytest")
def is_msamp_available(): return _is_package_available("msamp", "ms-amp")
def is_schedulefree_available(): return _is_package_available("schedulefree")
def is_transformer_engine_available(): return _is_package_available("transformer_engine", "transformer-engine")
def is_lomo_available(): return _is_package_available("lomo_optim")
def is_fp8_available(): return is_msamp_available() or is_transformer_engine_available()
def is_cuda_available():
    pytorch_nvml_based_cuda_check_previous_value = os.environ.get("PYTORCH_NVML_BASED_CUDA_CHECK")
    try:
        os.environ["PYTORCH_NVML_BASED_CUDA_CHECK"] = str(1)
        available = torch.cuda.is_available()
    finally:
        if pytorch_nvml_based_cuda_check_previous_value: os.environ["PYTORCH_NVML_BASED_CUDA_CHECK"] = pytorch_nvml_based_cuda_check_previous_value
        else: os.environ.pop("PYTORCH_NVML_BASED_CUDA_CHECK", None)
    return available
@lru_cache
def is_torch_xla_available(check_is_tpu=False, check_is_gpu=False):
    assert not (check_is_tpu and check_is_gpu), "The check_is_tpu and check_is_gpu cannot both be true."
    if not _torch_xla_available: return False
    elif check_is_gpu: return torch_xla.runtime.device_type() in ["GPU", "CUDA"]
    elif check_is_tpu: return torch_xla.runtime.device_type() == "TPU"
    return True
def is_deepspeed_available():
    if is_mlu_available(): return _is_package_available("deepspeed", metadata_name="deepspeed-mlu")
    return _is_package_available("deepspeed")
def is_pippy_available(): return is_torch_version(">=", "2.4.0")
def is_bf16_available(ignore_tpu=False):
    if is_torch_xla_available(check_is_tpu=True): return not ignore_tpu
    if is_cuda_available(): return torch.cuda.is_bf16_supported()
    if is_mps_available(): return False
    return True
def is_4bit_sapiens_available():
    package_exists = _is_package_available("sapiens_machine")
    if package_exists:
        sapiens_version = version.parse(importlib.metadata.version("sapiens_machine"))
        return compare_versions(sapiens_version, ">=", "0.39.0")
    return False
def is_8bit_sapiens_available():
    package_exists = _is_package_available("sapiens_machine")
    if package_exists:
        sapiens_version = version.parse(importlib.metadata.version("sapiens_machine"))
        return compare_versions(sapiens_version, ">=", "0.37.2")
    return False
def is_sapiens_available(): return _is_package_available("sapiens_machine")
def is_sapiens_machine_multi_backend_available():
    if not is_sapiens_available(): return False
    import sapiens_machine as sapiens
    return "multi_backend" in getattr(sapiens, "features", set())
def is_torchvision_available(): return _is_package_available("torchvision")
def is_megatron_lm_available():
    if str_to_bool(os.environ.get("SAPIENS_ACCELERATOR_USE_MEGATRON_LM", "False")) == 1:
        if importlib.util.find_spec("megatron") is not None:
            try:
                megatron_version = parse(importlib.metadata.version("megatron-core"))
                if compare_versions(megatron_version, "==", "0.5.0"): return importlib.util.find_spec(".data", "megatron")
            except Exception as e: return False
def is_transformers_available(): return _is_package_available("transformers")
def is_datasets_available(): return _is_package_available("datasets")
def is_peft_available(): return _is_package_available("peft")
def is_timm_available(): return _is_package_available("timm")
def is_triton_available(): return _is_package_available("triton")
def is_aim_available():
    package_exists = _is_package_available("aim")
    if package_exists:
        aim_version = version.parse(importlib.metadata.version("aim"))
        return compare_versions(aim_version, "<", "4.0.0")
    return False
def is_tensorboard_available(): return _is_package_available("tensorboard") or _is_package_available("tensorboardX")
def is_wandb_available(): return _is_package_available("wandb")
def is_comet_ml_available(): return _is_package_available("comet_ml")
def is_boto3_available(): return _is_package_available("boto3")
def is_rich_available():
    if _is_package_available("rich"): return parse_flag_from_env("SAPIENS_ACCELERATOR_ENABLE_RICH", False)
    return False
def is_sagemaker_available(): return _is_package_available("sagemaker")
def is_tqdm_available(): return _is_package_available("tqdm")
def is_clearml_available(): return _is_package_available("clearml")
def is_pandas_available(): return _is_package_available("pandas")
def is_mlflow_available():
    if _is_package_available("mlflow"): return True
    if importlib.util.find_spec("mlflow") is not None:
        try:
            _ = importlib.metadata.metadata("mlflow-skinny")
            return True
        except importlib.metadata.PackageNotFoundError: return False
    return False
def is_mps_available(min_version="1.12"): return is_torch_version(">=", min_version) and torch.backends.mps.is_available() and torch.backends.mps.is_built()
def is_ipex_available():
    def get_major_and_minor_from_version(full_version): return str(version.parse(full_version).major) + "." + str(version.parse(full_version).minor)
    _torch_version = importlib.metadata.version("torch")
    if importlib.util.find_spec("intel_extension_for_pytorch") is None: return False
    _ipex_version = "N/A"
    try: _ipex_version = importlib.metadata.version("intel_extension_for_pytorch")
    except importlib.metadata.PackageNotFoundError: return False
    torch_major_and_minor = get_major_and_minor_from_version(_torch_version)
    ipex_major_and_minor = get_major_and_minor_from_version(_ipex_version)
    if torch_major_and_minor != ipex_major_and_minor: return False
    return True
@lru_cache
def is_mlu_available(check_device=False):
    if importlib.util.find_spec("torch_mlu") is None: return False
    import torch_mlu
    if check_device:
        try:
            _ = torch.mlu.device_count()
            return torch.mlu.is_available()
        except RuntimeError: return False
    return hasattr(torch, "mlu") and torch.mlu.is_available()
@lru_cache
def is_musa_available(check_device=False):
    if importlib.util.find_spec("torch_musa") is None: return False
    import torch_musa
    if check_device:
        try:
            _ = torch.musa.device_count()
            return torch.musa.is_available()
        except RuntimeError: return False
    return hasattr(torch, "musa") and torch.musa.is_available()
@lru_cache
def is_npu_available(check_device=False):
    if importlib.util.find_spec("torch_npu") is None: return False
    import torch_npu
    if check_device:
        try:
            _ = torch.npu.device_count()
            return torch.npu.is_available()
        except RuntimeError: return False
    return hasattr(torch, "npu") and torch.npu.is_available()
@lru_cache
def is_xpu_available(check_device=False):
    if not parse_flag_from_env("SAPIENS_ACCELERATOR_USE_XPU", default=True): return False
    if is_ipex_available():
        if is_torch_version("<=", "1.12"): return False
        import intel_extension_for_pytorch
    else:
        if is_torch_version("<=", "2.3"): return False
    if check_device:
        try:
            _ = torch.xpu.device_count()
            return torch.xpu.is_available()
        except RuntimeError: return False
    return hasattr(torch, "xpu") and torch.xpu.is_available()
def is_dvclive_available(): return _is_package_available("dvclive")
def is_torchdata_available(): return _is_package_available("torchdata")
def is_torchdata_stateful_dataloader_available():
    package_exists = _is_package_available("torchdata")
    if package_exists:
        torchdata_version = version.parse(importlib.metadata.version("torchdata"))
        return compare_versions(torchdata_version, ">=", "0.8.0")
    return False
def deepspeed_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        from sapiens_accelerator.state import AcceleratorState
        from sapiens_accelerator.utils.dataclasses import DistributedType
        if AcceleratorState._shared_state != {} and AcceleratorState().distributed_type != DistributedType.DEEPSPEED: raise ValueError("DeepSpeed is not enabled, please make sure that an `Accelerator` is configured for `deepspeed` before calling this function.")
        return func(*args, **kwargs)
    return wrapper
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import math
from contextlib import suppress
from typing import Callable, List, Optional, Union
import torch
from torch.utils.data import BatchSampler, DataLoader, IterableDataset, RandomSampler
from .logging import get_logger
from .state import DistributedType, GradientState, PartialState, is_torch_xla_available
from .utils import (RNGType, broadcast, broadcast_object_list, concatenate, find_batch_size, get_data_structure, initialize_tensors, is_torch_version,
is_torchdata_stateful_dataloader_available, send_to_device, slice_tensors, synchronize_rng_states)
logger = get_logger(__name__)
_PYTORCH_DATALOADER_KWARGS = {"batch_size": 1, "shuffle": False, "sampler": None, "batch_sampler": None, "num_workers": 0, "collate_fn": None, "pin_memory": False,
"drop_last": False, "timeout": 0, "worker_init_fn": None, "multiprocessing_context": None, "generator": None, "prefetch_factor": 2, "persistent_workers": False}
_PYTORCH_DATALOADER_ADDITIONAL_KWARGS = {}
for v, additional_kwargs in _PYTORCH_DATALOADER_ADDITIONAL_KWARGS.items():
    if is_torch_version(">=", v): _PYTORCH_DATALOADER_KWARGS.update(additional_kwargs)
class SeedableRandomSampler(RandomSampler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epoch = 0
        self.initial_seed = torch.random.initial_seed()
    def __iter__(self):
        if self.generator is None:
            self.generator = torch.Generator()
            self.generator.manual_seed(self.initial_seed)
        seed = self.epoch + self.initial_seed
        self.generator.manual_seed(seed)
        yield from super().__iter__()
        self.set_epoch(self.epoch + 1)
    def set_epoch(self, epoch: int): self.epoch = epoch
class BatchSamplerShard(BatchSampler):
    def __init__(self, batch_sampler: BatchSampler, num_processes: int = 1, process_index: int = 0, split_batches: bool = False, even_batches: bool = True):
        if split_batches and batch_sampler.batch_size % num_processes != 0: raise ValueError(f"To use `BatchSamplerShard` in `split_batches` mode, the batch size ({batch_sampler.batch_size}) needs to be a round multiple of the number of processes ({num_processes}).")
        self.batch_sampler = batch_sampler
        self.num_processes = num_processes
        self.process_index = process_index
        self.split_batches = split_batches
        self.even_batches = even_batches
        self.batch_size = getattr(batch_sampler, "batch_size", None)
        self.drop_last = getattr(batch_sampler, "drop_last", False)
        if self.batch_size is None and self.even_batches: raise ValueError("You need to use `even_batches=False` when the batch sampler has no batch size. If you are not calling this method directly, set `accelerator.even_batches=False` instead.")
    @property
    def total_length(self): return len(self.batch_sampler)
    def __len__(self):
        if self.split_batches: return len(self.batch_sampler)
        if len(self.batch_sampler) % self.num_processes == 0: return len(self.batch_sampler) // self.num_processes
        length = len(self.batch_sampler) // self.num_processes
        if self.drop_last: return length
        elif self.even_batches: return length + 1
        else: return length + 1 if self.process_index < len(self.batch_sampler) % self.num_processes else length
    def __iter__(self): return self._iter_with_split() if self.split_batches else self._iter_with_no_split()
    def _iter_with_split(self):
        initial_data = []
        batch_length = self.batch_sampler.batch_size // self.num_processes
        for idx, batch in enumerate(self.batch_sampler):
            if idx == 0: initial_data = batch
            if len(batch) == self.batch_size: yield batch[batch_length * self.process_index : batch_length * (self.process_index + 1)]
        if not self.drop_last and len(initial_data) > 0 and len(batch) < self.batch_size:
            if not self.even_batches:
                if len(batch) > batch_length * self.process_index: yield batch[batch_length * self.process_index : batch_length * (self.process_index + 1)]
            else:
                while len(initial_data) < self.batch_size: initial_data += initial_data
                batch = batch + initial_data
                yield batch[batch_length * self.process_index : batch_length * (self.process_index + 1)]
    def _iter_with_no_split(self):
        initial_data = []
        batch_to_yield = []
        for idx, batch in enumerate(self.batch_sampler):
            if not self.drop_last and idx < self.num_processes: initial_data += batch
            if idx % self.num_processes == self.process_index: batch_to_yield = batch
            if idx % self.num_processes == self.num_processes - 1 and (self.batch_size is None or len(batch) == self.batch_size):
                yield batch_to_yield
                batch_to_yield = []
        if not self.drop_last and len(initial_data) > 0:
            if not self.even_batches:
                if len(batch_to_yield) > 0: yield batch_to_yield
            else:
                if len(batch_to_yield) == self.batch_size: yield batch_to_yield
                while len(initial_data) < self.num_processes * self.batch_size: initial_data += initial_data
                if len(batch) == self.batch_size:
                    batch = []
                    idx += 1
                cycle_index = 0
                while idx % self.num_processes != 0 or len(batch) > 0:
                    end_index = cycle_index + self.batch_size - len(batch)
                    batch += initial_data[cycle_index:end_index]
                    if idx % self.num_processes == self.process_index: yield batch
                    cycle_index = end_index
                    batch = []
                    idx += 1
class IterableDatasetShard(IterableDataset):
    def __init__(self, dataset: IterableDataset, batch_size: int = 1, drop_last: bool = False, num_processes: int = 1, process_index: int = 0, split_batches: bool = False):
        if split_batches and batch_size > 1 and batch_size % num_processes != 0: raise ValueError(f"To use `IterableDatasetShard` in `split_batches` mode, the batch size ({batch_size}) needs to be a round multiple of the number of processes ({num_processes}).")
        self.dataset = dataset
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.num_processes = num_processes
        self.process_index = process_index
        self.split_batches = split_batches
    def set_epoch(self, epoch):
        self.epoch = epoch
        if hasattr(self.dataset, "set_epoch"): self.dataset.set_epoch(epoch)
    def __len__(self):
        if self.drop_last: return (len(self.dataset) // (self.batch_size * self.num_processes)) * self.batch_size
        else: return math.ceil(len(self.dataset) / (self.batch_size * self.num_processes)) * self.batch_size
    def __iter__(self):
        if (not hasattr(self.dataset, "set_epoch") and hasattr(self.dataset, "generator") and isinstance(self.dataset.generator, torch.Generator)): self.dataset.generator.manual_seed(self.epoch)
        real_batch_size = self.batch_size if self.split_batches else (self.batch_size * self.num_processes)
        process_batch_size = (self.batch_size // self.num_processes) if self.split_batches else self.batch_size
        process_slice = range(self.process_index * process_batch_size, (self.process_index + 1) * process_batch_size)
        first_batch = None
        current_batch = []
        for element in self.dataset:
            current_batch.append(element)
            if len(current_batch) == real_batch_size:
                for i in process_slice: yield current_batch[i]
                if first_batch is None: first_batch = current_batch.copy()
                current_batch = []
        if not self.drop_last and len(current_batch) > 0:
            if first_batch is None: first_batch = current_batch.copy()
            while len(current_batch) < real_batch_size: current_batch += first_batch
            for i in process_slice: yield current_batch[i]
class DataLoaderStateMixin:
    def __init_subclass__(cls, **kwargs):
        cls.end_of_dataloader = False
        cls.remainder = -1
    def reset(self):
        self.end_of_dataloader = False
        self.remainder = -1
    def begin(self):
        self.reset()
        with suppress(Exception):
            if not self._drop_last:
                length = getattr(self.dataset, "total_dataset_length", len(self.dataset))
                self.remainder = length % self.total_batch_size
        self.gradient_state._add_dataloader(self)
    def end(self): self.gradient_state._remove_dataloader(self)
class DataLoaderAdapter:
    def __init__(self, dataset, use_stateful_dataloader=False, batch_sampler=None, **kwargs):
        self.use_stateful_dataloader = use_stateful_dataloader
        if is_torchdata_stateful_dataloader_available(): from torchdata.stateful_dataloader import StatefulDataLoader
        if use_stateful_dataloader and not is_torchdata_stateful_dataloader_available(): raise ImportError("StatefulDataLoader is not available. Please install torchdata version 0.8.0 or higher to use it.")
        if use_stateful_dataloader: self.base_dataloader = StatefulDataLoader(dataset, batch_sampler=batch_sampler, **kwargs)
        else: self.base_dataloader = DataLoader(dataset, batch_sampler=batch_sampler, **kwargs)
        if hasattr(self.base_dataloader, "state_dict"): self.dl_state_dict = self.base_dataloader.state_dict()
    def __getattr__(self, name):
        if name == "base_dataloader": raise AttributeError()
        return getattr(self.base_dataloader, name)
    def state_dict(self): return self.dl_state_dict
    def load_state_dict(self, state_dict): self.base_dataloader.load_state_dict(state_dict)
    @property
    def __class__(self): return self.base_dataloader.__class__
    def __len__(self): return len(self.base_dataloader)
    def adjust_state_dict_for_prefetch(self):
        if PartialState().distributed_type != DistributedType.NO:
            factor = PartialState().num_processes - 1
            if self.dl_state_dict["_sampler_iter_yielded"] > 0: self.dl_state_dict["_sampler_iter_yielded"] -= factor
            if self.dl_state_dict["_num_yielded"] > 0: self.dl_state_dict["_num_yielded"] -= factor
            if self.dl_state_dict["_index_sampler_state"] is not None:
                if ("samples_yielded" in self.dl_state_dict["_index_sampler_state"] and self.dl_state_dict["_index_sampler_state"]["samples_yielded"] > 0): self.dl_state_dict["_index_sampler_state"]["samples_yielded"] -= self.batch_size * factor
    def _update_state_dict(self):
        if hasattr(self.base_dataloader, "state_dict"):
            self.dl_state_dict = self.base_dataloader.state_dict()
            self.adjust_state_dict_for_prefetch()
            self.dl_state_dict["_iterator_finished"] = self.end_of_dataloader
class DataLoaderShard(DataLoaderAdapter, DataLoaderStateMixin):
    def __init__(self, dataset, device=None, rng_types=None, synchronized_generator=None, skip_batches=0, use_stateful_dataloader=False,
    _drop_last: bool = False, _non_blocking: bool = False, **kwargs):
        super().__init__(dataset, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
        self.device = device
        self.rng_types = rng_types
        self.synchronized_generator = synchronized_generator
        self.skip_batches = skip_batches
        self.gradient_state = GradientState()
        self._drop_last = _drop_last
        self._non_blocking = _non_blocking
        self.iteration = 0
    def __iter__(self):
        if self.rng_types is not None: synchronize_rng_states(self.rng_types, self.synchronized_generator)
        self.begin()
        self.set_epoch(self.iteration)
        dataloader_iter = self.base_dataloader.__iter__()
        try: current_batch = next(dataloader_iter)
        except StopIteration: yield
        batch_index = 0
        while True:
            try:
                if self.device is not None: current_batch = send_to_device(current_batch, self.device, non_blocking=self._non_blocking)
                self._update_state_dict()
                next_batch = next(dataloader_iter)
                if batch_index >= self.skip_batches: yield current_batch
                batch_index += 1
                current_batch = next_batch
            except StopIteration:
                self.end_of_dataloader = True
                self._update_state_dict()
                if batch_index >= self.skip_batches: yield current_batch
                break
        self.iteration += 1
        self.end()
    def __reduce__(self):
        args = super().__reduce__()
        return (DataLoaderShard, *args[1:])
    def set_epoch(self, epoch: int):
        if self.iteration != epoch: self.iteration = epoch
        if hasattr(self.batch_sampler, "sampler") and hasattr(self.batch_sampler.sampler, "set_epoch"): self.batch_sampler.sampler.set_epoch(epoch)
        elif hasattr(self.dataset, "set_epoch"): self.dataset.set_epoch(epoch)
    @property
    def total_batch_size(self):
        batch_sampler = self.sampler if isinstance(self.sampler, BatchSampler) else self.batch_sampler
        return (batch_sampler.batch_size if getattr(batch_sampler, "split_batches", False) else (batch_sampler.batch_size * getattr(batch_sampler, "num_processes", 1)))
    @property
    def total_dataset_length(self):
        if hasattr(self.dataset, "total_length"): return self.dataset.total_length
        else: return len(self.dataset)
    def get_sampler(self): return get_sampler(self)
    def set_sampler(self, sampler):
        sampler_is_batch_sampler = isinstance(self.sampler, BatchSampler)
        if sampler_is_batch_sampler: self.sampler.sampler = sampler
        else:
            self.batch_sampler.sampler = sampler
            if hasattr(self.batch_sampler, "batch_sampler"): self.batch_sampler.batch_sampler.sampler = sampler
if is_torch_xla_available():
    import torch_xla.distributed.parallel_loader as xpl
    class MpDeviceLoaderWrapper(xpl.MpDeviceLoader):
        def __init__(self, dataloader: DataLoaderShard, device: torch.device):
            super().__init__(dataloader, device)
            self._rng_types = self._loader.rng_types
            self._loader.rng_types = None
            self.device = device
        def __iter__(self):
            if self._rng_types is not None: synchronize_rng_states(self._rng_types, self._loader.synchronized_generator)
            return super().__iter__()
        def set_epoch(self, epoch: int):
            if hasattr(self.dataloader, "set_epoch"): self.dataloader.set_epoch(epoch)
        @property
        def total_batch_size(self): return self._loader.total_batch_size
        @property
        def total_dataset_length(self): return self._loader.total_dataset_length
        @property
        def batch_sampler(self): return self._loader.batch_sampler
        @property
        def dataloader(self): return self._loader
class DataLoaderDispatcher(DataLoaderAdapter, DataLoaderStateMixin):
    def __init__(self, dataset, split_batches: bool = False, skip_batches=0, use_stateful_dataloader=False, _drop_last: bool = False, _non_blocking: bool = False, slice_fn=None, **kwargs):
        shuffle = False
        if is_torch_version(">=", "1.11.0"):
            from torch.utils.data.datapipes.iter.combinatorics import ShufflerIterDataPipe
            if isinstance(dataset, ShufflerIterDataPipe): shuffle = dataset._shuffle_enabled
        super().__init__(dataset, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
        self.split_batches = split_batches
        if shuffle: torch.utils.data.graph_settings.apply_shuffle_settings(dataset, shuffle=shuffle)
        self.gradient_state = GradientState()
        self.state = PartialState()
        self._drop_last = _drop_last
        self._non_blocking = _non_blocking
        self.skip_batches = skip_batches
        self.slice_fn = slice_tensors if slice_fn is None else slice_fn
        self.iteration = 0
    def _fetch_batches(self, iterator):
        batches, batch = None, None
        if self.state.process_index == 0:
            try:
                if self.split_batches:
                    self._update_state_dict()
                    batch = next(iterator)
                else:
                    batches = []
                    for _ in range(self.state.num_processes):
                        self._update_state_dict()
                        batches.append(next(iterator))
                    try: batch = concatenate(batches, dim=0)
                    except RuntimeError as e: raise RuntimeError("You can't use batches of different size with `dispatch_batches=True` or when using an `IterableDataset`. Either pass `dispatch_batches=False` and have each process fetch its own batch or pass `split_batches=True`. By doing so, the main process will fetch a full batch and slice it into `num_processes` batches for each process.") from e
                batch_info = [get_data_structure(batch), False]
            except StopIteration: batch_info = [None, True]
        else: batch_info = [None, self._stop_iteration]
        broadcast_object_list(batch_info)
        self._stop_iteration = batch_info[1]
        if self._stop_iteration:
            if not self.split_batches and not self._drop_last:
                if self.state.process_index == 0 and len(batches) > 0:
                    batch = concatenate(batches, dim=0)
                    batch_info = [get_data_structure(batch), False]
                else: batch_info = [None, True]
                broadcast_object_list(batch_info)
        return batch, batch_info
    def __iter__(self):
        self.begin()
        self.set_epoch(self.iteration)
        main_iterator = None
        if is_torch_version(">=", "2.0.1"): main_iterator = self.base_dataloader.__iter__()
        elif self.state.process_index == 0: main_iterator = self.base_dataloader.__iter__()
        stop_iteration = False
        self._stop_iteration = False
        first_batch = None
        next_batch, next_batch_info = self._fetch_batches(main_iterator)
        batch_index = 0
        while not stop_iteration:
            batch, batch_info = next_batch, next_batch_info
            if self.state.process_index != 0: batch = initialize_tensors(batch_info[0])
            batch = send_to_device(batch, self.state.device, non_blocking=self._non_blocking)
            batch = broadcast(batch, from_process=0)
            if not self._drop_last and first_batch is None: first_batch = self.slice_fn(batch, slice(0, self.state.num_processes), process_index=self.state.process_index, num_processes=self.state.num_processes)
            if batch is None: raise ValueError(f"Batch does not contain any data (`{batch}`). At the end of all iterable data available before expected stop iteration.")
            observed_batch_size = find_batch_size(batch)
            batch_size = observed_batch_size // self.state.num_processes
            stop_iteration = self._stop_iteration
            if not stop_iteration:
                next_batch, next_batch_info = self._fetch_batches(main_iterator)
                if self._stop_iteration and next_batch_info[0] is None: stop_iteration = True
            if not self._drop_last and stop_iteration and observed_batch_size % self.state.num_processes != 0:
                batch = concatenate([batch, first_batch], dim=0)
                batch_size += 1
            data_slice = slice(self.state.process_index * batch_size, (self.state.process_index + 1) * batch_size)
            batch = self.slice_fn(batch, data_slice, process_index=self.state.process_index, num_processes=self.state.num_processes)
            if stop_iteration:
                self.end_of_dataloader = True
                self._update_state_dict()
                self.remainder = observed_batch_size
            if batch_index >= self.skip_batches: yield batch
            batch_index += 1
        self.iteration += 1
        self.end()
    def set_epoch(self, epoch: int):
        if self.iteration != epoch: self.iteration = epoch
        if hasattr(self.batch_sampler, "sampler") and hasattr(self.batch_sampler.sampler, "set_epoch"): self.batch_sampler.sampler.set_epoch(epoch)
        elif hasattr(self.dataset, "set_epoch"): self.dataset.set_epoch(epoch)
    def __len__(self):
        whole_length = len(self.base_dataloader)
        if self.split_batches: return whole_length
        elif self._drop_last: return whole_length // self.state.num_processes
        else: return math.ceil(whole_length / self.state.num_processes)
    def __reduce__(self):
        args = super().__reduce__()
        return (DataLoaderDispatcher, *args[1:])
    @property
    def total_batch_size(self): return (self.dataset.batch_size if self.split_batches else (self.dataset.batch_size * self.dataset.num_processes))
    @property
    def total_dataset_length(self): return len(self.dataset)
    def get_sampler(self): return get_sampler(self)
    def set_sampler(self, sampler):
        sampler_is_batch_sampler = isinstance(self.sampler, BatchSampler)
        if sampler_is_batch_sampler: self.sampler.sampler = sampler
        else:
            self.batch_sampler.sampler = sampler
            if hasattr(self.batch_sampler, "batch_sampler"): self.batch_sampler.batch_sampler.sampler = sampler
def get_sampler(dataloader):
    sampler_is_batch_sampler = isinstance(dataloader.sampler, BatchSampler)
    if sampler_is_batch_sampler: sampler = getattr(dataloader.sampler, "sampler", None)
    else: sampler = getattr(dataloader.batch_sampler, "sampler", None)
    return sampler
def prepare_data_loader(dataloader: DataLoader, device: Optional[torch.device] = None, num_processes: Optional[int] = None, process_index: Optional[int] = None,
split_batches: bool = False, put_on_device: bool = False, rng_types: Optional[List[Union[str, RNGType]]] = None, dispatch_batches: Optional[bool] = None,
even_batches: bool = True, slice_fn_for_dispatch: Optional[Callable] = None, use_seedable_sampler: bool = False, non_blocking: bool = False, use_stateful_dataloader: bool = False) -> DataLoader:
    if dispatch_batches is None:
        if not put_on_device: dispatch_batches = False
        else: dispatch_batches = isinstance(dataloader.dataset, IterableDataset)
    if dispatch_batches and not put_on_device: raise ValueError("Using `dispatch_batches=True` requires `put_on_device=True`.")
    state = PartialState()
    if num_processes is None: num_processes = state.num_processes
    if process_index is None: process_index = state.process_index
    if split_batches:
        if dataloader.batch_size is not None: batch_size_for_check = dataloader.batch_size
        else:
            if hasattr(dataloader.batch_sampler, "batch_size"): batch_size_for_check = dataloader.batch_sampler.batch_size
            else: raise ValueError(f"In order to use `split_batches==True` you must have a `batch_size` attribute either in the passed `dataloader` or `dataloader.batch_sampler` objects, and it has to return a natural number. Your `dataloader.batch_size` is None and `dataloader.batch_sampler` (`{type(dataloader.batch_sampler)}`) does not have the `batch_size` attribute set.")
        if batch_size_for_check > 1 and batch_size_for_check % num_processes != 0: raise ValueError(f"To use a `DataLoader` in `split_batches` mode, the batch size ({dataloader.batch_size}) needs to be a round multiple of the number of processes ({num_processes}).")
    new_dataset = dataloader.dataset
    new_batch_sampler = dataloader.batch_sampler if not isinstance(new_dataset, IterableDataset) else None
    sampler_is_batch_sampler = isinstance(dataloader.sampler, BatchSampler)
    synchronized_generator = None
    sampler = get_sampler(dataloader)
    if isinstance(sampler, RandomSampler) and use_seedable_sampler: sampler = SeedableRandomSampler(data_source=sampler.data_source, replacement=sampler.replacement, num_samples=sampler._num_samples, generator=getattr(sampler, "generator", torch.Generator()))
    if isinstance(dataloader.sampler, RandomSampler) and state.distributed_type == DistributedType.XLA:
        generator = torch.Generator().manual_seed(42)
        dataloader.generator = generator
        dataloader.sampler.generator = generator
    if (num_processes != 1 or state.distributed_type == DistributedType.MEGATRON_LM) and not dispatch_batches:
        if isinstance(new_dataset, IterableDataset):
            if getattr(dataloader.dataset, "generator", None) is not None: synchronized_generator = dataloader.dataset.generator
            new_dataset = IterableDatasetShard(new_dataset, batch_size=dataloader.batch_size, drop_last=dataloader.drop_last, num_processes=num_processes,
            process_index=process_index, split_batches=split_batches)
        else:
            if not use_seedable_sampler and hasattr(sampler, "generator"):
                if sampler.generator is None: sampler.generator = torch.Generator()
                synchronized_generator = sampler.generator
            batch_sampler = dataloader.sampler if sampler_is_batch_sampler else dataloader.batch_sampler
            new_batch_sampler = BatchSamplerShard(batch_sampler, num_processes=num_processes, process_index=process_index, split_batches=split_batches, even_batches=even_batches)
    ignore_kwargs = ["batch_size", "shuffle", "sampler", "batch_sampler", "drop_last"]
    if rng_types is not None and synchronized_generator is None and "generator" in rng_types: rng_types.remove("generator")
    kwargs = {k: getattr(dataloader, k, _PYTORCH_DATALOADER_KWARGS[k]) for k in _PYTORCH_DATALOADER_KWARGS if k not in ignore_kwargs}
    if new_batch_sampler is None:
        kwargs["drop_last"] = dataloader.drop_last
        kwargs["batch_size"] = (dataloader.batch_size // num_processes if split_batches and not dispatch_batches else dataloader.batch_size)
    if dispatch_batches:
        kwargs.pop("generator")
        dataloader = DataLoaderDispatcher(new_dataset, split_batches=split_batches, batch_sampler=new_batch_sampler, _drop_last=dataloader.drop_last,
        _non_blocking=non_blocking, slice_fn=slice_fn_for_dispatch, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
    elif sampler_is_batch_sampler: dataloader = DataLoaderShard(new_dataset, device=device if put_on_device and state.distributed_type != DistributedType.XLA else None,
    sampler=new_batch_sampler, batch_size=dataloader.batch_size, rng_types=rng_types, _drop_last=dataloader.drop_last, _non_blocking=non_blocking,
    synchronized_generator=synchronized_generator, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
    else: dataloader = DataLoaderShard(new_dataset, device=device if put_on_device and state.distributed_type != DistributedType.XLA else None,
    batch_sampler=new_batch_sampler, rng_types=rng_types, synchronized_generator=synchronized_generator, _drop_last=dataloader.drop_last,
    _non_blocking=non_blocking, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
    if isinstance(sampler, SeedableRandomSampler) and use_seedable_sampler: dataloader.set_sampler(sampler)
    if state.distributed_type == DistributedType.XLA: return MpDeviceLoaderWrapper(dataloader, device)
    return dataloader
class SkipBatchSampler(BatchSampler):
    def __init__(self, batch_sampler, skip_batches=0):
        self.batch_sampler = batch_sampler
        self.skip_batches = skip_batches
    def __iter__(self):
        for index, samples in enumerate(self.batch_sampler):
            if index >= self.skip_batches: yield samples
    @property
    def total_length(self): return len(self.batch_sampler)
    def __len__(self): return len(self.batch_sampler) - self.skip_batches
class SkipDataLoader(DataLoaderAdapter, DataLoaderStateMixin):
    def __init__(self, dataset, skip_batches=0, use_stateful_dataloader=False, **kwargs):
        super().__init__(dataset, use_stateful_dataloader=use_stateful_dataloader, **kwargs)
        self.skip_batches = skip_batches
        self.gradient_state = GradientState()
    def __iter__(self):
        self.begin()
        for index, batch in enumerate(self.base_dataloader.__iter__()):
            if index >= self.skip_batches:
                self._update_state_dict()
                yield batch
        self.end()
    def __len__(self): return len(self.base_dataloader) - self.skip_batches
    def __reduce__(self):
        args = super().__reduce__()
        return (SkipDataLoader, *args[1:])
def skip_first_batches(dataloader, num_batches=0):
    state = PartialState()
    if state.distributed_type == DistributedType.XLA:
        device = dataloader.device
        dataloader = dataloader.dataloader
    dataset = dataloader.dataset
    sampler_is_batch_sampler = False
    if isinstance(dataset, IterableDataset): new_batch_sampler = None
    else:
        sampler_is_batch_sampler = isinstance(dataloader.sampler, BatchSampler)
        batch_sampler = dataloader.sampler if sampler_is_batch_sampler else dataloader.batch_sampler
        new_batch_sampler = SkipBatchSampler(batch_sampler, skip_batches=num_batches)
    ignore_kwargs = ["batch_size", "shuffle", "sampler", "batch_sampler", "drop_last"]
    kwargs = {k: getattr(dataloader, k, _PYTORCH_DATALOADER_KWARGS[k]) for k in _PYTORCH_DATALOADER_KWARGS if k not in ignore_kwargs}
    if new_batch_sampler is None:
        kwargs["drop_last"] = dataloader.drop_last
        kwargs["batch_size"] = dataloader.batch_size
    if isinstance(dataloader, DataLoaderDispatcher):
        if new_batch_sampler is None: kwargs["skip_batches"] = num_batches
        dataloader = DataLoaderDispatcher(dataset, split_batches=dataloader.split_batches, batch_sampler=new_batch_sampler, _drop_last=dataloader._drop_last, **kwargs)
    elif isinstance(dataloader, DataLoaderShard):
        if new_batch_sampler is None: kwargs["skip_batches"] = num_batches
        elif sampler_is_batch_sampler:
            kwargs["sampler"] = new_batch_sampler
            kwargs["batch_size"] = dataloader.batch_size
        else: kwargs["batch_sampler"] = new_batch_sampler
        dataloader = DataLoaderShard(dataset, device=dataloader.device, rng_types=dataloader.rng_types, synchronized_generator=dataloader.synchronized_generator, **kwargs)
    else:
        if new_batch_sampler is None: dataloader = SkipDataLoader(dataset, skip_batches=num_batches, **kwargs)
        else: dataloader = DataLoader(dataset, batch_sampler=new_batch_sampler, **kwargs)
    if state.distributed_type == DistributedType.XLA: dataloader = MpDeviceLoaderWrapper(dataloader, device)
    return dataloader
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import importlib
import logging
import os
import subprocess
import sys
from pathlib import Path
import psutil
import torch
from sapiens_accelerator.commands.config import default_config_file, load_config_from_file
from sapiens_accelerator.commands.config.config_args import SageMakerConfig
from sapiens_accelerator.commands.config.config_utils import DYNAMO_BACKENDS
from sapiens_accelerator.commands.utils import CustomArgumentParser
from sapiens_accelerator.state import get_int_from_env
from sapiens_accelerator.utils import (ComputeEnvironment, DistributedType, PrepareForLaunch, _filter_args, check_cuda_p2p_ib_support, convert_dict_to_env_variables,
is_bf16_available, is_deepspeed_available, is_mlu_available, is_musa_available, is_npu_available, is_rich_available, is_sagemaker_available, is_torch_version,
is_torch_xla_available, is_xpu_available, patch_environment, prepare_deepspeed_cmd_env, prepare_multi_gpu_env, prepare_sagemager_args_inputs,
prepare_simple_launcher_cmd_env, prepare_tpu, str_to_bool)
from sapiens_accelerator.utils.constants import DEEPSPEED_MULTINODE_LAUNCHERS, TORCH_DYNAMO_MODES
if is_rich_available():
    from rich import get_console
    from rich.logging import RichHandler
    FORMAT = "%(message)s"
    logging.basicConfig(format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
logger = logging.getLogger(__name__)
options_to_group = {"multi_gpu": "Distributed GPUs", "tpu": "TPU", "use_deepspeed": "DeepSpeed Arguments", "use_fsdp": "FSDP Arguments", "use_megatron_lm": "Megatron-LM Arguments", "fp8_backend": "FP8 Arguments"}
def clean_option(option):
    if "fp8_backend" in option: option = "--fp8_backend"
    if option.startswith("--"): return option[2:].replace("-", "_")
class CustomHelpFormatter(argparse.HelpFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titles = ["Hardware Selection Arguments", "Resource Selection Arguments", "Training Paradigm Arguments", "positional arguments", "optional arguments"]
    def add_argument(self, action: argparse.Action):
        if "sapiens_accelerator" in sys.argv[0] and "launch" in sys.argv[1:]: args = sys.argv[2:]
        else: args = sys.argv[1:]
        if len(args) > 1:
            args = list(map(clean_option, args))
            used_platforms = [arg for arg in args if arg in options_to_group.keys()]
            used_titles = [options_to_group[o] for o in used_platforms]
            if action.container.title not in self.titles + used_titles: action.help = argparse.SUPPRESS
            elif action.container.title == "Hardware Selection Arguments":
                if set(action.option_strings).isdisjoint(set(args)): action.help = argparse.SUPPRESS
                else: action.help = action.help + " (currently selected)"
            elif action.container.title == "Training Paradigm Arguments":
                if set(action.option_strings).isdisjoint(set(args)): action.help = argparse.SUPPRESS
                else: action.help = action.help + " (currently selected)"
        action.option_strings = [s for s in action.option_strings if "-" not in s[2:]]
        super().add_argument(action)
    def end_section(self):
        if len(self._current_section.items) < 2:
            self._current_section.items = []
            self._current_section.heading = ""
        super().end_section()
def launch_command_parser(subparsers=None):
    description = "Launch a python script in a distributed scenario. Arguments can be passed in with either hyphens (`--num-processes=2`) or underscores (`--num_processes=2`)"
    if subparsers is not None: parser = subparsers.add_parser("launch", description=description, add_help=False, allow_abbrev=False, formatter_class=CustomHelpFormatter)
    else: parser = CustomArgumentParser("SapiensAccelerator launch command", description=description, add_help=False, allow_abbrev=False, formatter_class=CustomHelpFormatter)
    if subparsers is not None: parser.set_defaults(func=launch_command)
    return parser
def simple_launcher(args):
    cmd, current_env = prepare_simple_launcher_cmd_env(args)
    process = subprocess.Popen(cmd, env=current_env)
    process.wait()
    if process.returncode != 0:
        if not args.quiet: raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)
        else: sys.exit(1)
def multi_gpu_launcher(args):
    import torch.distributed.run as distrib_run
    current_env = prepare_multi_gpu_env(args)
    if not check_cuda_p2p_ib_support():
        message = "Using RTX 4000 series which doesn't support faster communication speedups. Ensuring P2P and IB communications are disabled."
        warn = False
        if "NCCL_P2P_DISABLE" not in current_env:
            current_env["NCCL_P2P_DISABLE"] = "1"
            warn = True
        if "NCCL_IB_DISABLE" not in current_env:
            current_env["NCCL_IB_DISABLE"] = "1"
            warn = True
    debug = getattr(args, "debug", False)
    args = _filter_args(args, distrib_run.get_args_parser(), ["--training_script", args.training_script, "--training_script_args", args.training_script_args])
    with patch_environment(**current_env):
        try: distrib_run.run(args)
        except Exception:
            if is_rich_available() and debug:
                console = get_console()
                console.print("\n[bold red]Using --debug, `torch.distributed` Stack Trace:[/bold red]")
                console.print_exception(suppress=[__file__], show_locals=False)
            else: raise
def deepspeed_launcher(args):
    import torch.distributed.run as distrib_run
    if not is_deepspeed_available(): raise ImportError("DeepSpeed is not installed => run `pip3 install deepspeed` or build it from source.")
    else: from deepspeed.launcher.runner import DEEPSPEED_ENVIRONMENT_NAME
    cmd, current_env = prepare_deepspeed_cmd_env(args)
    if not check_cuda_p2p_ib_support():
        message = "Using RTX 4000 series which doesn't support faster communication speedups. Ensuring P2P and IB communications are disabled."
        warn = False
        if "NCCL_P2P_DISABLE" not in current_env:
            current_env["NCCL_P2P_DISABLE"] = "1"
            warn = True
        if "NCCL_IB_DISABLE" not in current_env:
            current_env["NCCL_IB_DISABLE"] = "1"
            warn = True
    if args.num_machines > 1 and args.deepspeed_multinode_launcher != DEEPSPEED_MULTINODE_LAUNCHERS[1]:
        with open(DEEPSPEED_ENVIRONMENT_NAME, "a") as f:
            valid_env_items = convert_dict_to_env_variables(current_env)
            if len(valid_env_items) > 1: f.writelines(valid_env_items)
        process = subprocess.Popen(cmd, env=current_env)
        process.wait()
        if process.returncode != 0:
            if not args.quiet: raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)
            else: sys.exit(1)
    else:
        debug = getattr(args, "debug", False)
        args = _filter_args(args, distrib_run.get_args_parser(), ["--training_script", args.training_script, "--training_script_args", args.training_script_args])
        with patch_environment(**current_env):
            try: distrib_run.run(args)
            except Exception:
                if is_rich_available() and debug:
                    console = get_console()
                    console.print("\n[bold red]Using --debug, `torch.distributed` Stack Trace:[/bold red]")
                    console.print_exception(suppress=[__file__], show_locals=False)
                else: raise
def tpu_launcher(args):
    import torch_xla.distributed.xla_multiprocessing as xmp
    if args.no_python: raise ValueError("--no_python cannot be used with TPU launcher")
    args, current_env = prepare_tpu(args, {})
    if args.module: mod_name = args.training_script
    else:
        script_path = Path(args.training_script)
        sys.path.append(str(script_path.parent.resolve()))
        mod_name = script_path.stem
    mod = importlib.import_module(mod_name)
    if not hasattr(mod, args.main_training_function): raise ValueError(f"Your training script should have a function named {args.main_training_function}, or you should pass a different value to `--main_training_function`.")
    sys.argv = [mod.__file__] + args.training_script_args
    main_function = getattr(mod, args.main_training_function)
    with patch_environment(**current_env): xmp.spawn(PrepareForLaunch(main_function), args=(), nprocs=args.num_processes)
def tpu_pod_launcher(args):
    from torch_xla.distributed import xla_dist
    current_env = {}
    args, current_env = prepare_tpu(args, current_env, True)
    debug = getattr(args, "debug", False)
    training_script = args.training_script
    training_script_args = args.training_script_args
    new_args = _filter_args(args, xla_dist.get_args_parser(), ["--tpu", args.tpu_name, "--positional", "", "--restart-tpuvm-pod-server"])
    if args.tpu_use_sudo: new_cmd = ["sudo"]
    else: new_cmd = []
    new_cmd += ["sapiens_accelerator-launch", "--tpu", "--no_tpu_cluster", "--num_machines", "1", "--mixed_precision", "no", "--dynamo_backend", "no", "--num_processes",
    str(args.num_processes), "--main_training_function", str(args.main_training_function), training_script] + training_script_args
    new_args.positional = new_cmd
    bad_flags = ""
    for arg in vars(new_args):
        if arg.startswith("docker_"):
            value = getattr(new_args, arg)
            if value != "" and value is not None: bad_flags += f'{arg}="{value}"\n'
    if bad_flags != "": raise ValueError(f"Docker containers are not supported for TPU pod launcher currently, please remove the following flags:\n{bad_flags}")
    new_args.env = [f"{k}={v}" for k, v in current_env.items()]
    new_args.env.append("SAPIENS_ACCELERATOR_IN_TPU_POD=1")
    try: xla_dist.resolve_and_execute(new_args)
    except Exception:
        if is_rich_available() and debug:
            console = get_console()
            console.print("\n[bold red]Using --debug, `torch_xla.xla_dist` Stack Trace:[/bold red]")
            console.print_exception(suppress=[__file__], show_locals=False)
        else: raise
def sagemaker_launcher(sagemaker_config: SageMakerConfig, args):
    if not is_sagemaker_available(): raise ImportError("Please install sagemaker to be able to launch training on Amazon SageMaker with `pip install sapiens_accelerator[sagemaker]`")
    if args.module or args.no_python: raise ValueError("SageMaker requires a python training script file and cannot be used with --module or --no_python")
    from sagemaker.huggingface import HuggingFace
    args, sagemaker_inputs = prepare_sagemager_args_inputs(sagemaker_config, args)
    huggingface_estimator = HuggingFace(**args)
    huggingface_estimator.fit(inputs=sagemaker_inputs)
    print(f"You can find your model data at: {huggingface_estimator.model_data}")
def _validate_launch_command(args):
    if sum([args.multi_gpu, args.cpu, args.tpu, args.use_deepspeed, args.use_fsdp]) > 1: raise ValueError("You can only use one of `--cpu`, `--multi_gpu`, `--tpu`, `--use_deepspeed`, `--use_fsdp` at a time.")
    if args.multi_gpu and (args.num_processes is not None) and (args.num_processes < 2): raise ValueError("You need to use at least 2 processes to use `--multi_gpu`.")
    defaults = None
    warned = []
    mp_from_config_flag = False
    if args.config_file is not None or os.path.isfile(default_config_file) and not args.cpu:
        defaults = load_config_from_file(args.config_file)
        if (not args.multi_gpu and not args.tpu and not args.tpu_use_cluster and not args.use_deepspeed and not args.use_fsdp and not args.use_megatron_lm):
            args.use_deepspeed = defaults.distributed_type == DistributedType.DEEPSPEED
            args.multi_gpu = (True if defaults.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_XPU) else False)
            args.tpu = defaults.distributed_type == DistributedType.XLA
            args.use_fsdp = defaults.distributed_type == DistributedType.FSDP
            args.use_megatron_lm = defaults.distributed_type == DistributedType.MEGATRON_LM
            args.tpu_use_cluster = defaults.tpu_use_cluster if args.tpu else False
        if args.gpu_ids is None:
            if defaults.gpu_ids is not None: args.gpu_ids = defaults.gpu_ids
            else: args.gpu_ids = "all"
        if args.multi_gpu and args.num_machines is None: args.num_machines = defaults.num_machines
        if len(args.gpu_ids.split(",")) < 2 and (args.gpu_ids != "all") and args.multi_gpu and args.num_machines <= 1: raise ValueError("Less than two GPU ids were configured and tried to run on on multiple GPUs. Please ensure at least two are specified for `--gpu_ids`, or use `--gpu_ids='all'`.")
        if defaults.compute_environment == ComputeEnvironment.LOCAL_MACHINE:
            for name, attr in defaults.__dict__.items():
                if isinstance(attr, dict):
                    for k in defaults.deepspeed_config: setattr(args, k, defaults.deepspeed_config[k])
                    for k in defaults.fsdp_config:
                        arg_to_set = k
                        if "fsdp" not in arg_to_set: arg_to_set = "fsdp_" + arg_to_set
                        setattr(args, arg_to_set, defaults.fsdp_config[k])
                    for k in defaults.megatron_lm_config: setattr(args, k, defaults.megatron_lm_config[k])
                    for k in defaults.dynamo_config: setattr(args, k, defaults.dynamo_config[k])
                    for k in defaults.ipex_config: setattr(args, k, defaults.ipex_config[k])
                    for k in defaults.mpirun_config: setattr(args, k, defaults.mpirun_config[k])
                    continue
                if (name not in ["compute_environment", "mixed_precision", "distributed_type"] and getattr(args, name, None) is None): setattr(args, name, attr)
        if not args.debug: args.debug = defaults.debug
        if not args.mixed_precision:
            if defaults.mixed_precision is None: args.mixed_precision = "no"
            else:
                args.mixed_precision = defaults.mixed_precision
                mp_from_config_flag = True
        else:
            if args.use_cpu or (args.use_xpu and torch.xpu.is_available()): native_amp = is_torch_version(">=", "1.10")
            else: native_amp = is_bf16_available(True)
            if (args.mixed_precision == "bf16" and not native_amp and not (args.tpu and is_torch_xla_available(check_is_tpu=True))): raise ValueError("bf16 mixed precision requires PyTorch >= 1.10 and a supported device.")
        if args.dynamo_backend is None: args.dynamo_backend = "no"
        if args.num_processes == -1: raise ValueError("You need to manually pass in `--num_processes` using this config yaml.")
    else:
        if args.num_processes is None:
            if args.use_xpu and is_xpu_available(): args.num_processes = torch.xpu.device_count()
            elif is_mlu_available(): args.num_processes = torch.mlu.device_count()
            elif is_musa_available(): args.num_processes = torch.musa.device_count()
            elif is_npu_available(): args.num_processes = torch.npu.device_count()
            else: args.num_processes = torch.cuda.device_count()
            warned.append(f"\t`--num_processes` was set to a value of `{args.num_processes}`")
        if args.debug is None: args.debug = False
        if (not args.multi_gpu and args.num_processes > 1 and ((args.use_xpu and is_xpu_available() and torch.xpu.device_count() > 1) or (is_mlu_available() and torch.mlu.device_count() > 1)
        or (is_musa_available() and torch.musa.device_count() > 1) or (is_npu_available() and torch.npu.device_count() > 1) or (torch.cuda.device_count() > 1))):
            warned.append("\t\tMore than one GPU was found, enabling multi-GPU training.\n\t\tIf this was unintended please pass in `--num_processes=1`.")
            args.multi_gpu = True
        if args.num_machines is None:
            warned.append("\t`--num_machines` was set to a value of `1`")
            args.num_machines = 1
        if args.mixed_precision is None:
            warned.append("\t`--mixed_precision` was set to a value of `'no'`")
            args.mixed_precision = "no"
        if not hasattr(args, "use_cpu"): args.use_cpu = args.cpu
        if args.dynamo_backend is None:
            warned.append("\t`--dynamo_backend` was set to a value of `'no'`")
            args.dynamo_backend = "no"
    is_aws_env_disabled = defaults is None or (defaults is not None and defaults.compute_environment != ComputeEnvironment.AMAZON_SAGEMAKER)
    if is_aws_env_disabled and args.num_cpu_threads_per_process is None:
        args.num_cpu_threads_per_process = get_int_from_env(["OMP_NUM_THREADS"], 1)
        if args.use_cpu and args.num_processes >= 1 and get_int_from_env(["OMP_NUM_THREADS"], 0) == 0:
            local_size = get_int_from_env(["MPI_LOCALNRANKS", "OMPI_COMM_WORLD_LOCAL_SIZE", "MV2_COMM_WORLD_LOCAL_SIZE"], max(int(args.num_processes / args.num_machines), 1))
            threads_per_process = int(psutil.cpu_count(logical=False) / local_size)
            if threads_per_process > 1:
                args.num_cpu_threads_per_process = threads_per_process
                warned.append(f"\t`--num_cpu_threads_per_process` was set to `{args.num_cpu_threads_per_process}` to improve out-of-box performance when training on CPUs")
    return args, defaults, mp_from_config_flag
def launch_command(args):
    args, defaults, mp_from_config_flag = _validate_launch_command(args)
    if args.use_deepspeed and not args.cpu:
        args.deepspeed_fields_from_sapiens_accelerator_config = list(defaults.deepspeed_config.keys()) if defaults else []
        if mp_from_config_flag: args.deepspeed_fields_from_sapiens_accelerator_config.append("mixed_precision")
        args.deepspeed_fields_from_sapiens_accelerator_config = ",".join(args.deepspeed_fields_from_sapiens_accelerator_config)
        deepspeed_launcher(args)
    elif args.use_fsdp and not args.cpu: multi_gpu_launcher(args)
    elif args.use_megatron_lm and not args.cpu: multi_gpu_launcher(args)
    elif args.multi_gpu and not args.cpu: multi_gpu_launcher(args)
    elif args.tpu and not args.cpu:
        if args.tpu_use_cluster: tpu_pod_launcher(args)
        else: tpu_launcher(args)
    elif defaults is not None and defaults.compute_environment == ComputeEnvironment.AMAZON_SAGEMAKER: sagemaker_launcher(defaults, args)
    else: simple_launcher(args)
def main():
    parser = launch_command_parser()
    args = parser.parse_args()
    launch_command(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from sapiens_accelerator.commands.config import get_config_parser
from sapiens_accelerator.commands.env import env_command_parser
from sapiens_accelerator.commands.estimate import estimate_command_parser
from sapiens_accelerator.commands.launch import launch_command_parser
from sapiens_accelerator.commands.merge import merge_command_parser
from sapiens_accelerator.commands.test import test_command_parser
from sapiens_accelerator.commands.tpu import tpu_command_parser
from sapiens_accelerator.commands.utils import CustomArgumentParser
def main():
    parser = CustomArgumentParser("SapiensAccelerator CLI tool", usage="sapiens_accelerator <command> [<args>]", allow_abbrev=False)
    subparsers = parser.add_subparsers(help="sapiens_accelerator command helpers")
    get_config_parser(subparsers=subparsers)
    estimate_command_parser(subparsers=subparsers)
    env_command_parser(subparsers=subparsers)
    launch_command_parser(subparsers=subparsers)
    merge_command_parser(subparsers=subparsers)
    tpu_command_parser(subparsers=subparsers)
    test_command_parser(subparsers=subparsers)
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)
    args.func(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import os
import platform
import subprocess
import numpy as np
import psutil
import torch
from sapiens_accelerator import __version__ as version
from sapiens_accelerator.commands.config import default_config_file, load_config_from_file
from ..utils import is_mlu_available, is_musa_available, is_npu_available, is_xpu_available
def env_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("env")
    else: parser = argparse.ArgumentParser("SapiensAccelerator env command")
    parser.add_argument("--config_file", default=None, help="The config file to use for the default values in the launching script.")
    if subparsers is not None: parser.set_defaults(func=env_command)
    return parser
def env_command(args):
    pt_version = torch.__version__
    pt_cuda_available = torch.cuda.is_available()
    pt_xpu_available = is_xpu_available()
    pt_mlu_available = is_mlu_available()
    pt_musa_available = is_musa_available()
    pt_npu_available = is_npu_available()
    sapiens_accelerator_config = "Not found"
    if args.config_file is not None or os.path.isfile(default_config_file): sapiens_accelerator_config = load_config_from_file(args.config_file).to_dict()
    command = None
    bash_location = "Not found"
    if os.name == "nt": command = ["where", "sapiens_accelerator"]
    elif os.name == "posix": command = ["which", "sapiens_accelerator"]
    if command is not None: bash_location = subprocess.check_output(command, text=True, stderr=subprocess.STDOUT).strip()
    info = {"`SapiensAccelerator` version": version, "Platform": platform.platform(), "`sapiens_accelerator` bash location": bash_location, "Python version": platform.python_version(),
    "Numpy version": np.__version__, "PyTorch version (GPU?)": f"{pt_version} ({pt_cuda_available})", "PyTorch XPU available": str(pt_xpu_available),
    "PyTorch NPU available": str(pt_npu_available), "PyTorch MLU available": str(pt_mlu_available), "PyTorch MUSA available": str(pt_musa_available),
    "System RAM": f"{psutil.virtual_memory().total / 1024 ** 3:.2f} GB"}
    if pt_cuda_available: info["GPU type"] = torch.cuda.get_device_name()
    if pt_mlu_available: info["MLU type"] = torch.mlu.get_device_name()
    if pt_npu_available: info["CANN version"] = torch.version.cann
    sapiens_accelerator_config_str = ("\n".join([f"\t- {prop}: {val}" for prop, val in sapiens_accelerator_config.items()]) if isinstance(sapiens_accelerator_config, dict) else f"\t{sapiens_accelerator_config}")
    info["`SapiensAccelerator` configs"] = sapiens_accelerator_config
    return info
def main() -> int:
    parser = env_command_parser()
    args = parser.parse_args()
    env_command(args)
    return 0
if __name__ == "__main__": raise SystemExit(main())
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import json
import os
import evaluate
import torch
from datasets import load_dataset
from torch.optim import AdamW
from torch.utils.data import DataLoader
from transformers import AutoModelForSequenceClassification, AutoTokenizer, get_linear_schedule_with_warmup, set_seed
from sapiens_accelerator import Accelerator, DistributedType
from sapiens_accelerator.utils.deepspeed import DummyOptim, DummyScheduler
MAX_GPU_BATCH_SIZE = 16
EVAL_BATCH_SIZE = 32
def get_dataloaders(accelerator: Accelerator, batch_size: int = 16, model_name: str = "bert-base-cased"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    datasets = load_dataset("glue", "mrpc")
    def tokenize_function(examples):
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None)
        return outputs
    tokenized_datasets = datasets.map(tokenize_function, batched=True, remove_columns=["idx", "sentence1", "sentence2"], load_from_cache_file=False)
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    def collate_fn(examples):
        if accelerator.distributed_type == DistributedType.XLA: return tokenizer.pad(examples, padding="max_length", max_length=128, return_tensors="pt")
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")
    train_dataloader = DataLoader(tokenized_datasets["train"], shuffle=True, collate_fn=collate_fn, batch_size=batch_size)
    eval_dataloader = DataLoader(tokenized_datasets["validation"], shuffle=False, collate_fn=collate_fn, batch_size=EVAL_BATCH_SIZE)
    return train_dataloader, eval_dataloader
def evaluation_loop(accelerator, model, eval_dataloader, metric):
    model.eval()
    samples_seen = 0
    for step, batch in enumerate(eval_dataloader):
        batch.to(accelerator.device)
        with torch.no_grad(): outputs = model(**batch)
        predictions = outputs.logits.argmax(dim=-1)
        predictions, references = accelerator.gather((predictions, batch["labels"]))
        if accelerator.use_distributed:
            if step == len(eval_dataloader) - 1:
                predictions = predictions[: len(eval_dataloader.dataset) - samples_seen]
                references = references[: len(eval_dataloader.dataset) - samples_seen]
            else: samples_seen += references.shape[0]
        metric.add_batch(predictions=predictions, references=references)
    eval_metric = metric.compute()
    return eval_metric["accuracy"]
def training_function(config, args):
    accelerator = Accelerator()
    lr = config["lr"]
    num_epochs = int(config["num_epochs"])
    seed = int(config["seed"])
    batch_size = int(config["batch_size"])
    model_name = args.model_name_or_path
    set_seed(seed)
    train_dataloader, eval_dataloader = get_dataloaders(accelerator, batch_size, model_name)
    model = AutoModelForSequenceClassification.from_pretrained(model_name, return_dict=True)
    optimizer_cls = (AdamW if accelerator.state.deepspeed_plugin is None or "optimizer" not in accelerator.state.deepspeed_plugin.deepspeed_config else DummyOptim)
    optimizer = optimizer_cls(params=model.parameters(), lr=lr)
    if accelerator.state.deepspeed_plugin is not None: gradient_accumulation_steps = accelerator.state.deepspeed_plugin.deepspeed_config["gradient_accumulation_steps"]
    else: gradient_accumulation_steps = 1
    max_training_steps = (len(train_dataloader) * num_epochs) // gradient_accumulation_steps
    if (accelerator.state.deepspeed_plugin is None or "scheduler" not in accelerator.state.deepspeed_plugin.deepspeed_config): lr_scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
    else: lr_scheduler = DummyScheduler(optimizer, total_num_steps=max_training_steps, warmup_num_steps=0)
    model, optimizer, train_dataloader, eval_dataloader, lr_scheduler = accelerator.prepare(model, optimizer, train_dataloader, eval_dataloader, lr_scheduler)
    overall_step = 0
    starting_epoch = 0
    metric = evaluate.load("glue", "mrpc")
    ending_epoch = num_epochs
    if args.partial_train_epoch is not None: ending_epoch = args.partial_train_epoch
    if args.resume_from_checkpoint:
        accelerator.load_state(args.resume_from_checkpoint)
        epoch_string = args.resume_from_checkpoint.split("epoch_")[1]
        state_epoch_num = ""
        for char in epoch_string:
            if char.isdigit(): state_epoch_num += char
            else: break
        starting_epoch = int(state_epoch_num) + 1
        accuracy = evaluation_loop(accelerator, model, eval_dataloader, metric)
        accelerator.print("resumed checkpoint performance:", accuracy)
        accelerator.print("resumed checkpoint's scheduler's lr:", lr_scheduler.get_lr()[0])
        accelerator.print("resumed optimizers's lr:", optimizer.param_groups[0]["lr"])
        with open(os.path.join(args.output_dir, f"state_{starting_epoch - 1}.json")) as f:
            resumed_state = json.load(f)
            assert resumed_state["accuracy"] == accuracy, "Accuracy mismatch, loading from checkpoint failed"
            assert (resumed_state["lr"] == lr_scheduler.get_lr()[0]), "Scheduler learning rate mismatch, loading from checkpoint failed"
            assert (resumed_state["optimizer_lr"] == optimizer.param_groups[0]["lr"]), "Optimizer learning rate mismatch, loading from checkpoint failed"
            assert resumed_state["epoch"] == starting_epoch - 1, "Epoch mismatch, loading from checkpoint failed"
            return
    state = {}
    for epoch in range(starting_epoch, ending_epoch):
        model.train()
        for step, batch in enumerate(train_dataloader):
            outputs = model(**batch)
            loss = outputs.loss
            loss = loss / gradient_accumulation_steps
            accelerator.backward(loss)
            if step % gradient_accumulation_steps == 0:
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
            overall_step += 1
        output_dir = f"epoch_{epoch}"
        output_dir = os.path.join(args.output_dir, output_dir)
        accelerator.save_state(output_dir)
        accuracy = evaluation_loop(accelerator, model, eval_dataloader, metric)
        state["accuracy"] = accuracy
        state["lr"] = lr_scheduler.get_lr()[0]
        state["optimizer_lr"] = optimizer.param_groups[0]["lr"]
        state["epoch"] = epoch
        state["step"] = overall_step
        accelerator.print(f"epoch {epoch}:", state)
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            with open(os.path.join(args.output_dir, f"state_{epoch}.json"), "w") as f: json.dump(state, f)
    accelerator.end_training()
def main():
    parser = argparse.ArgumentParser(description="Simple example of training script tracking peak GPU memory usage.")
    parser.add_argument("--model_name_or_path", type=str, default="bert-base-cased", help="", required=False)
    parser.add_argument("--output_dir", type=str, default=".", help="Optional save directory where all checkpoint folders will be stored. Default is the current working directory.")
    parser.add_argument("--resume_from_checkpoint", type=str, default=None, help="If the training should continue from a checkpoint folder.")
    parser.add_argument("--partial_train_epoch", type=int, default=None, help="If passed, the training will stop after this number of epochs.")
    parser.add_argument("--num_epochs", type=int, default=2, help="Number of train epochs.")
    args = parser.parse_args()
    config = {"lr": 2e-5, "num_epochs": args.num_epochs, "seed": 42, "batch_size": 16}
    training_function(config, args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import importlib.metadata
from typing import Union
from packaging.version import Version, parse
from .constants import STR_OPERATION_TO_FUNC
torch_version = parse(importlib.metadata.version("torch"))
def compare_versions(library_or_version: Union[str, Version], operation: str, requirement_version: str):
    if operation not in STR_OPERATION_TO_FUNC.keys(): raise ValueError(f"`operation` must be one of {list(STR_OPERATION_TO_FUNC.keys())}, received {operation}")
    operation = STR_OPERATION_TO_FUNC[operation]
    if isinstance(library_or_version, str): library_or_version = parse(importlib.metadata.version(library_or_version))
    return operation(library_or_version, parse(requirement_version))
def is_torch_version(operation: str, version: str): return compare_versions(torch_version, operation, version)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import os
import subprocess
from packaging.version import Version, parse
from sapiens_accelerator.commands.config.config_args import default_config_file, load_config_from_file
_description = "Run commands across TPU VMs for initial setup before running `sapiens_accelerator launch`."
def tpu_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("tpu-config", description=_description)
    else: parser = argparse.ArgumentParser("SapiensAccelerator tpu-config command", description=_description)
    config_args = parser.add_argument_group("Config Arguments", "Arguments that can be configured through `sapiens_accelerator config`.")
    config_args.add_argument("--config_file", type=str, default=None, help="Path to the config file to use for sapiens_accelerator.")
    config_args.add_argument("--tpu_name", default=None, help="The name of the TPU to use. If not specified, will use the TPU specified in the config file.")
    config_args.add_argument("--tpu_zone", default=None, help="The zone of the TPU to use. If not specified, will use the zone specified in the config file.")
    pod_args = parser.add_argument_group("TPU Arguments", "Arguments for options ran inside the TPU.")
    pod_args.add_argument("--use_alpha", action="store_true", help="Whether to use `gcloud alpha` when running the TPU training script instead of `gcloud`.")
    pod_args.add_argument("--command_file", default=None, help="The path to the file containing the commands to run on the pod on startup.")
    pod_args.add_argument("--command", action="append", nargs="+", help="A command to run on the pod. Can be passed multiple times.")
    pod_args.add_argument("--install_sapiens_accelerator", action="store_true", help="Whether to install sapiens_accelerator on the pod. Defaults to False.")
    pod_args.add_argument("--sapiens_accelerator_version", default="latest", help="The version of sapiens_accelerator to install on the pod. If not specified, will use the latest pypi version. Specify 'dev' to install from GitHub.")
    pod_args.add_argument("--debug", action="store_true", help="If set, will print the command that would be run instead of running it.")
    if subparsers is not None: parser.set_defaults(func=tpu_command_launcher)
    return parser
def tpu_command_launcher(args):
    defaults = None
    if args.config_file is not None or os.path.isfile(default_config_file):
        defaults = load_config_from_file(args.config_file)
        if not args.command_file and defaults.command_file is not None and not args.command: args.command_file = defaults.command_file
        if not args.command and defaults.commands is not None: args.command = defaults.commands
        if not args.tpu_name: args.tpu_name = defaults.tpu_name
        if not args.tpu_zone: args.tpu_zone = defaults.tpu_zone
    if args.sapiens_accelerator_version == "latest": args.sapiens_accelerator_version = "sapiens_accelerator -U"
    elif isinstance(parse(args.sapiens_accelerator_version), Version): args.sapiens_accelerator_version = f"sapiens_accelerator=={args.sapiens_accelerator_version}"
    if not args.command_file and not args.command: raise ValueError("You must specify either a command file or a command to run on the pod.")
    if args.command_file:
        with open(args.command_file) as f: args.command = [f.read().splitlines()]
    if isinstance(args.command[0], list): args.command = [line for cmd in args.command for line in cmd]
    new_cmd = ["cd /usr/share"]
    if args.install_sapiens_accelerator: new_cmd += [f"pip install {args.sapiens_accelerator_version}"]
    new_cmd += args.command
    args.command = "; ".join(new_cmd)
    cmd = ["gcloud"]
    if args.use_alpha: cmd += ["alpha"]
    cmd += ["compute", "tpus", "tpu-vm", "ssh", args.tpu_name, "--zone", args.tpu_zone, "--command", args.command, "--worker", "all"]
    if args.debug: return
    subprocess.run(cmd)
def main():
    parser = tpu_command_parser()
    args = parser.parse_args()
    tpu_command_launcher(args)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from copy import deepcopy
import torch
import torch.nn.functional as F
from torch.optim import AdamW
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader
from sapiens_accelerator.accelerator import Accelerator, DataLoaderConfiguration, GradientAccumulationPlugin
from sapiens_accelerator.state import GradientState
from sapiens_accelerator.test_utils import RegressionDataset, RegressionModel
from sapiens_accelerator.utils import DistributedType, set_seed
def check_model_parameters(model_a, model_b, did_step, iteration, **kwargs):
    for param, grad_param in zip(model_a.parameters(), model_b.parameters()):
        if not param.requires_grad: continue
        if not did_step: assert (torch.allclose(param.grad, grad_param.grad, **kwargs) is False), f"Gradients in sync when they should not be at iteration {iteration}:\nmodel_a grad ({param.grad}) == model_b grad ({grad_param.grad})"
        else: assert (torch.allclose(param.grad, grad_param.grad, **kwargs) is True), f"Gradients not in sync when they should be at iteration {iteration}:\nmodel_a grad ({param.grad}) != model_b grad ({grad_param.grad})"
def step_model(model, input, target, accelerator, do_backward=True):
    model.train()
    output = model(input)
    loss = F.mse_loss(output, target.to(output.device))
    if not do_backward:
        loss /= accelerator.gradient_accumulation_steps
        loss.backward()
    else: accelerator.backward(loss)
def get_training_setup(accelerator, sched=False):
    set_seed(42)
    model = RegressionModel()
    ddp_model = deepcopy(model)
    dset = RegressionDataset(length=80)
    dataloader = DataLoader(dset, batch_size=16)
    model.to(accelerator.device)
    if sched:
        opt = AdamW(params=model.parameters(), lr=1e-3)
        ddp_opt = AdamW(params=ddp_model.parameters(), lr=1e-3)
        sched = LambdaLR(opt, lr_lambda=lambda epoch: epoch**0.65)
        ddp_sched = LambdaLR(ddp_opt, lr_lambda=lambda epoch: epoch**0.65)
    if sched: ddp_model, ddp_opt, ddp_sched, dataloader = accelerator.prepare(ddp_model, ddp_opt, ddp_sched, dataloader)
    else: ddp_model, dataloader = accelerator.prepare(ddp_model, dataloader)
    if sched: return (model, opt, sched, dataloader, ddp_model, ddp_opt, ddp_sched)
    return model, ddp_model, dataloader
def test_noop_sync(accelerator):
    model, ddp_model, dataloader = get_training_setup(accelerator)
    ddp_input, ddp_target = next(iter(dataloader)).values()
    for iteration in range(3):
        input, target = accelerator.gather((ddp_input, ddp_target))
        input, target = input.to(accelerator.device), target.to(accelerator.device)
        step_model(model, input, target, accelerator)
        if iteration % 2 == 0:
            with accelerator.no_sync(ddp_model): step_model(ddp_model, ddp_input, ddp_target, accelerator)
        else: step_model(ddp_model, ddp_input, ddp_target, accelerator)
        check_model_parameters(model, ddp_model, True, iteration)
        for param, ddp_param in zip(model.parameters(), ddp_model.parameters()):
            if not param.requires_grad: continue
            assert torch.allclose(param.grad, ddp_param.grad), f"Gradients not in sync when they should be:\nModel grad ({param.grad}) != DDP grad ({ddp_param.grad})"
        torch.manual_seed(1337 + iteration)
        ddp_input = ddp_input[torch.randperm(len(ddp_input))]
def test_distributed_sync(accelerator):
    model, ddp_model, dataloader = get_training_setup(accelerator)
    ddp_input, ddp_target = next(iter(dataloader)).values()
    for iteration in range(3):
        input, target = accelerator.gather((ddp_input, ddp_target))
        input, target = input.to(accelerator.device), target.to(accelerator.device)
        step_model(model, input, target, accelerator)
        if iteration % 2 == 0:
            with accelerator.no_sync(ddp_model): step_model(ddp_model, ddp_input, ddp_target, accelerator)
        else: step_model(ddp_model, ddp_input, ddp_target, accelerator)
        for param, ddp_param in zip(model.parameters(), ddp_model.parameters()):
            if not param.requires_grad: continue
            if iteration % 2 == 0: assert (torch.allclose(param.grad, ddp_param.grad) is False), f"Gradients in sync when they should not be:\nModel grad ({param.grad}) == DDP grad ({ddp_param.grad})"
            else: assert (torch.allclose(param.grad, ddp_param.grad) is True), f"Gradients not in sync when they should be:\nModel grad ({param.grad}) != DDP grad ({ddp_param.grad})"
        torch.manual_seed(1337 + iteration)
        ddp_input = ddp_input[torch.randperm(len(ddp_input))]
def test_distributed_sync_multiple_fwd(accelerator):
    model, ddp_model, dataloader = get_training_setup(accelerator)
    losses = []
    num_iterations = 3
    for iteration in range(num_iterations):
        ddp_input, ddp_target = next(iter(dataloader)).values()
        input, target = accelerator.gather((ddp_input, ddp_target))
        input, target = input.to(accelerator.device), target.to(accelerator.device)
        step_model(model, input, target, accelerator)
        with accelerator.no_sync(ddp_model):
            ddp_output = ddp_model(ddp_input)
            loss = F.mse_loss(ddp_output, ddp_target.to(ddp_output.device))
            losses.append(loss)
    for iteration in range(num_iterations):
        loss = losses[iteration]
        if iteration < num_iterations - 1:
            accelerator.backward(loss)
            for param, ddp_param in zip(model.parameters(), ddp_model.parameters()):
                if not param.requires_grad: continue
                assert (torch.allclose(param.grad, ddp_param.grad) is False), f"Gradients in sync when they should not be:\nModel grad ({param.grad}) == DDP grad ({ddp_param.grad})"
        else:
            with accelerator.trigger_sync_in_backward(ddp_model): accelerator.backward(loss)
            for param, ddp_param in zip(model.parameters(), ddp_model.parameters()):
                if not param.requires_grad: continue
                assert (torch.allclose(param.grad, ddp_param.grad) is True), f"Gradients not in sync when they should be:\nModel grad ({param.grad}) != DDP grad ({ddp_param.grad})"
def test_gradient_accumulation(split_batches=False, dispatch_batches=False, sync_each_batch=False):
    gradient_accumulation_plugin = GradientAccumulationPlugin(num_steps=2, sync_each_batch=sync_each_batch)
    dataloader_config = DataLoaderConfiguration(split_batches=split_batches, dispatch_batches=dispatch_batches)
    accelerator = Accelerator(dataloader_config=dataloader_config, gradient_accumulation_plugin=gradient_accumulation_plugin)
    model, ddp_model, dataloader = get_training_setup(accelerator)
    for iteration, batch in enumerate(dataloader):
        ddp_input, ddp_target = batch.values()
        input, target = accelerator.gather((ddp_input, ddp_target))
        input, target = input.to(accelerator.device), target.to(accelerator.device)
        step_model(model, input, target, accelerator, False)
        with accelerator.accumulate(ddp_model): step_model(ddp_model, ddp_input, ddp_target, accelerator)
        for param, ddp_param in zip(model.parameters(), ddp_model.parameters()):
            if not param.requires_grad: continue
            if ((iteration + 1) % 2 == 0) or (iteration == len(dataloader) - 1) or sync_each_batch:
                assert (torch.allclose(param.grad, ddp_param.grad) is True), f"Gradients not in sync when they should be at iteration {iteration}:\nModel grad ({param.grad}) != DDP grad ({ddp_param.grad})"
            else: assert (torch.allclose(param.grad, ddp_param.grad) is False), f"Gradients in sync when they should not be at iteration {iteration}:\nModel grad ({param.grad}) == DDP grad ({ddp_param.grad})"
        torch.manual_seed(1337 + iteration)
        ddp_input = ddp_input[torch.randperm(len(ddp_input))]
    GradientState._reset_state()
def test_gradient_accumulation_with_opt_and_scheduler(split_batches=False, dispatch_batches=False, sync_each_batch=False):
    gradient_accumulation_plugin = GradientAccumulationPlugin(num_steps=2, sync_each_batch=sync_each_batch)
    dataloader_config = DataLoaderConfiguration(split_batches=split_batches, dispatch_batches=dispatch_batches)
    accelerator = Accelerator(dataloader_config=dataloader_config, gradient_accumulation_plugin=gradient_accumulation_plugin)
    model, opt, sched, dataloader, ddp_model, ddp_opt, ddp_sched = get_training_setup(accelerator, True)
    for iteration, batch in enumerate(dataloader):
        ddp_input, ddp_target = batch.values()
        input, target = accelerator.gather((ddp_input, ddp_target))
        input, target = input.to(accelerator.device), target.to(accelerator.device)
        model.train()
        ddp_model.train()
        step_model(model, input, target, accelerator, False)
        opt.step()
        if ((iteration + 1) % 2 == 0) or ((iteration + 1) == len(dataloader)):
            if split_batches: sched.step()
            else:
                for _ in range(accelerator.num_processes): sched.step()
        with accelerator.accumulate(ddp_model):
            step_model(ddp_model, ddp_input, ddp_target, accelerator)
            ddp_opt.step()
            ddp_sched.step()
        assert (opt.param_groups[0]["lr"] == ddp_opt.param_groups[0]["lr"]), f'Learning rates found in each optimizer did not align\nopt: {opt.param_groups[0]["lr"]}\nDDP opt: {ddp_opt.param_groups[0]["lr"]}\n'
        did_step = (((iteration + 1) % 2) == 0) or ((iteration + 1) == len(dataloader))
        if accelerator.num_processes > 1: check_model_parameters(model, ddp_model, did_step or sync_each_batch, iteration, rtol=1e-3)
        if did_step: opt.zero_grad()
        ddp_opt.zero_grad()
        torch.manual_seed(1337 + iteration)
    GradientState._reset_state()
def test_dataloader_break():
    accelerator = Accelerator()
    first_dset = RegressionDataset(length=80)
    first_dataloader = DataLoader(first_dset, batch_size=16)
    second_dset = RegressionDataset(length=96)
    second_dataloader = DataLoader(second_dset, batch_size=16)
    first_dataloader, second_dataloader = accelerator.prepare(first_dataloader, second_dataloader)
    assert accelerator.gradient_state.active_dataloader is None
    for iteration, _ in enumerate(first_dataloader):
        assert id(accelerator.gradient_state.active_dataloader) == id(first_dataloader)
        if iteration < len(first_dataloader) - 1:
            assert not accelerator.gradient_state.end_of_dataloader
            if iteration == 1:
                for batch_num, _ in enumerate(second_dataloader):
                    assert id(accelerator.gradient_state.active_dataloader) == id(second_dataloader)
                    if batch_num < len(second_dataloader) - 1: assert not accelerator.gradient_state.end_of_dataloader
                    else: assert accelerator.gradient_state.end_of_dataloader
        else: assert accelerator.gradient_state.end_of_dataloader
    assert accelerator.gradient_state.active_dataloader is None
def main():
    accelerator = Accelerator()
    state = accelerator.state
    if state.distributed_type != DistributedType.XLA: test_dataloader_break()
    if state.distributed_type == DistributedType.NO: test_noop_sync(accelerator)
    if state.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_CPU):
        test_distributed_sync(accelerator)
        test_distributed_sync_multiple_fwd(accelerator)
    if state.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA):
        for split_batch in [True, False]:
            for dispatch_batches in [True, False]:
                for sync_each_batch in [True, False]: test_gradient_accumulation(split_batch, dispatch_batches, sync_each_batch)
    test_gradient_accumulation_with_opt_and_scheduler()
    if state.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA):
        for split_batch in [True, False]:
            for dispatch_batches in [True, False]:
                for sync_each_batch in [True, False]:
                    if not split_batch and not dispatch_batches and not sync_each_batch: continue
                    test_gradient_accumulation_with_opt_and_scheduler(split_batch, dispatch_batches, sync_each_batch)
    state.destroy_process_group()
def _mp_fn(index): main()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
